@@ -30,21 +30,17 @@ my class Any { # declared in BOOTSTRAP
     proto method DELETE-KEY(|) is nodal { * }
     multi method DELETE-KEY(Any:U: $) { Nil }
     multi method DELETE-KEY(Any:D: $) {
-        fail "Can not remove values from a {self.^name}";
+        Failure.new("Can not remove values from a {self.^name}")
     }
 
     proto method DELETE-POS(|) is nodal { * }
     multi method DELETE-POS(Any:U: $pos) { Nil }
     multi method DELETE-POS(Any:D: $pos) {
-        fail "Can not remove elements from a {self.^name}";
+        Failure.new("Can not remove elements from a {self.^name}")
     }
     multi method DELETE-POS(**@indices) {
         my $final := @indices.pop;
-        my $target := self;
-        for @indices {
-            $target := $target.AT-POS($_);
-        }
-        $target.DELETE-POS($final);
+        Rakudo::Internals.WALK-AT-POS(self,@indices).DELETE-POS($final)
     }
 
     method cache() { self.list }
@@ -75,6 +71,9 @@ my class Any { # declared in BOOTSTRAP
     proto method Hash(|) is nodal { * }
     multi method Hash() { self.hash.Hash }
 
+    proto method Map(|) is nodal { * }
+    multi method Map() { self.hash.Map }
+
     proto method elems(|) is nodal { * }
     multi method elems(Any:U:) { 1 }
     multi method elems(Any:D:) { self.list.elems }
@@ -227,11 +226,7 @@ my class Any { # declared in BOOTSTRAP
     }
     multi method EXISTS-POS(**@indices) {
         my $final := @indices.pop;
-        my $target := self;
-        for @indices {
-            $target := $target.AT-POS($_);
-        }
-        $target.EXISTS-POS($final);
+        Rakudo::Internals.WALK-AT-POS(self,@indices).EXISTS-POS($final)
     }
 
     proto method AT-POS(|) is nodal {*}
@@ -248,30 +243,30 @@ my class Any { # declared in BOOTSTRAP
         $v
     }
     multi method AT-POS(Any:U: Num:D \pos) is raw {
-        fail X::Item.new(aggregate => self, index => pos)
-          if nqp::isnanorinf(pos);
-        self.AT-POS(nqp::unbox_i(pos.Int));
+        nqp::isnanorinf(pos)
+          ?? Failure.new(X::Item.new(aggregate => self, index => pos))
+          !! self.AT-POS(nqp::unbox_i(pos.Int))
     }
     multi method AT-POS(Any:U: Any:D \pos) is raw {
         self.AT-POS(nqp::unbox_i(pos.Int));
     }
 
     multi method AT-POS(Any:D: int \pos) is raw {
-        fail X::OutOfRange.new(
-          :what($*INDEX // 'Index'), :got(pos), :range<0..0>)
-            unless nqp::not_i(pos);
-        self;
+        pos
+          ?? Failure.new(X::OutOfRange.new(
+               :what($*INDEX // 'Index'), :got(pos), :range<0..0>))
+          !! self
     }
     multi method AT-POS(Any:D: Int:D \pos) is raw {
-        fail X::OutOfRange.new(
-          :what($*INDEX // 'Index'), :got(pos), :range<0..0>)
-            if pos != 0;
-        self;
+        pos
+          ?? Failure.new(X::OutOfRange.new(
+               :what($*INDEX // 'Index'), :got(pos), :range<0..0>))
+          !! self
     }
     multi method AT-POS(Any:D: Num:D \pos) is raw {
-        fail X::Item.new(aggregate => self, index => pos)
-          if nqp::isnanorinf(pos);
-        self.AT-POS(nqp::unbox_i(pos.Int));
+        nqp::isnanorinf(pos)
+          ?? Failure.new(X::Item.new(aggregate => self, index => pos))
+          !! self.AT-POS(nqp::unbox_i(pos.Int))
     }
     multi method AT-POS(Any:D: Any:D \pos) is raw {
         self.AT-POS(nqp::unbox_i(pos.Int));
@@ -280,30 +275,27 @@ my class Any { # declared in BOOTSTRAP
         die "Cannot use '{pos.^name}' as an index";
     }
     multi method AT-POS(**@indices) is raw {
-        my $result := self;
-        for @indices {
-            $result := $result.AT-POS($_);
-        }
-        $result
+        my $final := @indices.pop;
+        Rakudo::Internals.WALK-AT-POS(self,@indices).AT-POS($final)
     }
 
     proto method ZEN-POS(|) { * }
     multi method ZEN-POS(*%unexpected) {
         %unexpected
-          ?? fail X::Adverb.new(
+          ?? Failure.new(X::Adverb.new(
                :what('[] slice'),
                :source(try { self.VAR.name } // self.WHAT.perl),
-               :unexpected(%unexpected.keys))
+               :unexpected(%unexpected.keys)))
           !! self
     }
 
     proto method ZEN-KEY(|) { * }
     multi method ZEN-KEY(*%unexpected) {
         %unexpected
-          ?? fail X::Adverb.new(
+          ?? Failure.new(X::Adverb.new(
                :what('{} slice'),
                :source(try { self.VAR.name } // self.WHAT.perl),
-               :unexpected(%unexpected.keys))
+               :unexpected(%unexpected.keys)))
           !! self
     }
 
@@ -319,9 +311,9 @@ my class Any { # declared in BOOTSTRAP
         self.AT-POS(pos) = assignee;                    # defer < 0 check
     }
     multi method ASSIGN-POS(Any:D: Num:D \pos, Mu \assignee) {
-        fail X::Item.new(aggregate => self, index => pos)
-          if nqp::isnanorinf(pos);
-        self.AT-POS(nqp::unbox_i(pos.Int)) = assignee;  # defer < 0 check
+        nqp::isnanorinf(pos)
+          ?? Failure.new(X::Item.new(aggregate => self, index => pos))
+          !! self.AT-POS(nqp::unbox_i(pos.Int)) = assignee;  # defer < 0 check
     }
     multi method ASSIGN-POS(Any:D: Any:D \pos, Mu \assignee) {
         self.AT-POS(nqp::unbox_i(pos.Int)) = assignee;  # defer < 0 check
@@ -332,41 +324,38 @@ my class Any { # declared in BOOTSTRAP
     multi method ASSIGN-POS(**@indices) {
         my \value := @indices.pop;
         my $final := @indices.pop;
-        my $target := self;
-        for @indices {
-            $target := $target.AT-POS($_);
-        }
-        $target.ASSIGN-POS($final, value)
+        Rakudo::Internals.WALK-AT-POS(self,@indices).ASSIGN-POS($final,value)
     }
 
     proto method BIND-POS(|) { * }
     multi method BIND-POS(Any:D: **@indices is raw) is raw {
+# looks like Array.pop doesn't really return a bindable container
+#        my \value := @indices.pop;
+#        my $final := @indices.pop;
+#        Rakudo::Internals.WALK-AT-POS(self,@indices).BIND-POS($final,value)
+
         my int $elems = @indices.elems;
-        my \value := @indices.AT-POS($elems - 1);
-        my $final := @indices.AT-POS($elems - 2);
+        my \value  := @indices.AT-POS(--$elems);
+        my $final  := @indices.AT-POS(--$elems);
         my $target := self;
-        my int $i = 0;
-        while $i < $elems - 2 {
-            $target := $target.AT-POS(@indices.AT-POS($i));
-            $i = $i + 1;
-        }
+        my int $i = -1;
+        $target := $target.AT-POS(@indices.AT-POS($i))
+          while nqp::islt_i(++$i,$elems);
         $target.BIND-POS($final, value)
     }
 
-    method all() is nodal { Junction.new(self.list, :type<all>) }
-    method any() is nodal { Junction.new(self.list, :type<any>) }
-    method one() is nodal { Junction.new(self.list, :type<one>) }
-    method none() is nodal { Junction.new(self.list, :type<none>) }
+    method all()  is nodal { Junction.new("all", self) }
+    method any()  is nodal { Junction.new("any", self) }
+    method one()  is nodal { Junction.new("one", self) }
+    method none() is nodal { Junction.new("none",self) }
 
     # internals
     proto method AT-KEY(|) is nodal { * }
     multi method AT-KEY(Any:D: $key) is raw {
-        if self ~~ Associative {
-            fail "Associative indexing implementation missing from type {self.WHAT.perl}";
-        }
-        else {
-            fail "Type {self.WHAT.perl} does not support associative indexing.";
-        }
+        Failure.new( self ~~ Associative
+          ?? "Associative indexing implementation missing from type {self.WHAT.perl}"
+          !! "Type {self.WHAT.perl} does not support associative indexing."
+        )
     }
     multi method AT-KEY(Any:U \SELF: $key) is raw {
         nqp::bindattr(my $v, Scalar, '$!whence',
@@ -377,7 +366,7 @@ my class Any { # declared in BOOTSTRAP
 
     proto method BIND-KEY(|) is nodal { * }
     multi method BIND-KEY(Any:D: \k, \v) is raw {
-        fail X::Bind.new(target => self.^name);
+        Failure.new(X::Bind.new(target => self.^name))
     }
     multi method BIND-KEY(Any:U \SELF: $key, $BIND ) is raw {
         SELF = Hash.new;
@@ -415,8 +404,10 @@ my class Any { # declared in BOOTSTRAP
         my \iter = self.iterator;
         my $sum = 0;
         my Mu $value;
-        $sum = $sum + $value
-          until ($value := iter.pull-one) =:= IterationEnd;
+        nqp::until(
+          nqp::eqaddr(($value := iter.pull-one),IterationEnd),
+          ($sum = $sum + $value)
+        );
         $sum;
     }
 }
@@ -425,8 +416,11 @@ Metamodel::ClassHOW.exclude_parent(Any);
 # builtin ops
 proto sub infix:<===>(Mu $?, Mu $?) is pure { * }
 multi sub infix:<===>($?)    { Bool::True }
-multi sub infix:<===>($a, $b) {
-    nqp::p6bool(nqp::iseq_s(nqp::unbox_s($a.WHICH), nqp::unbox_s($b.WHICH)))
+multi sub infix:<===>(\a, \b) {
+    nqp::p6bool(
+      nqp::eqaddr(a.WHAT,b.WHAT)
+      && nqp::iseq_s(nqp::unbox_s(a.WHICH), nqp::unbox_s(b.WHICH))
+    )
 }
 
 proto sub infix:<before>(Mu $?, Mu $?)  is pure { * }
@@ -514,12 +508,12 @@ sub SLICE_HUH(\SELF, @nogo, %d, %adv) {
         }
     }
 
-    fail X::Adverb.new(
+    Failure.new(X::Adverb.new(
       :what<slice>,
       :source(try { SELF.VAR.name } // SELF.WHAT.perl),
       :unexpected(%d.keys),
       :nogo(@nogo),
-    );
+    ))
 } #SLICE_HUH
 
 sub DELETEKEY(Mu \d, str $key) {
@@ -529,24 +523,19 @@ sub DELETEKEY(Mu \d, str $key) {
         $value;
     }
     else {
-        Mu;
+        Nil;
     }
 } #DELETEKEY
 
-sub dd(**@p, *%n) {
-    if @p || %n {
-        for @p -> $var {
-            my $name := $var.VAR.?name;
+sub dd(|) {
+    my Mu $args := nqp::p6argvmarray();
+    if nqp::elems($args) {
+        while $args {
+            my $var  := nqp::shift($args);
+            my $name := try $var.VAR.?name;
             my $type := $var.WHAT.^name;
             my $what := $var.?is-lazy
-              ?? $var[^10].perl.chop ~ "...Inf)"
-              !! $var.perl;
-            note $name ?? "$type $name = $what" !! $what;
-        }
-        for %n.kv -> $name, $var {
-            my $type := $var.WHAT.^name;
-            my $what := $var.?is-lazy
-              ?? $var[^10].perl.chop ~ "...Inf)"
+              ?? $var[^10].perl.chop ~ "... (lazy list)"
               !! $var.perl;
             note $name ?? "$type $name = $what" !! $what;
         }
