@@ -1,9 +1,7 @@
 my class X::TypeCheck { ... };
 my class X::Subscript::Negative { ... };
-my class X::IllegalOnFixedDimensionArray { ... };
 my class X::NotEnoughDimensions { ... };
 my class X::Assignment::ArrayShapeMismatch { ... };
-my class X::Assignment::ToShaped { ... };
 
 # An Array is a List that ensures every item added to it is in a Scalar
 # container. It also supports push, pop, shift, unshift, splice, BIND-POS,
@@ -42,6 +40,146 @@ my class Array { # declared in BOOTSTRAP
         }
     }
 
+    method iterator(Array:D:) {
+
+        # something to iterate over in the future
+        if nqp::getattr(self,List,'$!todo').DEFINITE {
+            self!ensure-allocated;
+            class :: does Iterator {
+                has int $!i;
+                has $!array;
+                has $!reified;
+                has $!todo;
+                has $!descriptor;
+
+                method !SET-SELF(\array) {
+                    $!i           = -1;
+                    $!array      := array;
+                    $!reified    := nqp::getattr(array, List,  '$!reified');
+                    $!todo       := nqp::getattr(array, List,  '$!todo');
+                    $!descriptor := nqp::getattr(array, Array, '$!descriptor');
+                    self
+                }
+                method new(\array) { nqp::create(self)!SET-SELF(array) }
+
+                method pull-one() is raw {
+                    nqp::ifnull(
+                      nqp::atpos($!reified,$!i = nqp::add_i($!i,1)),
+                      nqp::islt_i($!i,nqp::elems($!reified))
+                        ?? self!found-hole
+                        !! $!todo.DEFINITE
+                          ?? nqp::islt_i($!i,$!todo.reify-at-least(nqp::add_i($!i,1)))
+                            ?? nqp::atpos($!reified,$!i) # cannot be nqp::null
+                            !! self!done
+                          !! IterationEnd
+                    )
+                }
+                method !found-hole() {
+                   nqp::p6bindattrinvres(
+                     (my \v := nqp::p6scalarfromdesc($!descriptor)),
+                     Scalar,
+                     '$!whence',
+                     -> { nqp::bindpos($!reified,$!i,v) }
+                   )
+                }
+                method !done() is raw {
+                    $!todo := nqp::bindattr($!array,List,'$!todo',Mu);
+                    IterationEnd
+                }
+
+                method push-until-lazy($target) {
+                    if $!todo.DEFINITE {
+                        my int $elems = $!todo.reify-until-lazy;
+                        nqp::while(   # doesn't sink
+                          nqp::islt_i($!i = nqp::add_i($!i,1),$elems),
+                          $target.push(nqp::atpos($!reified,$!i))
+                        );
+                        nqp::if(
+                          $!todo.fully-reified,
+                          self!done,
+                          nqp::stmts(
+                            ($!i = nqp::sub_i($elems,1)),
+                            Mu
+                          )
+                        )
+                    }
+                    else {
+                        my int $elems = nqp::elems($!reified);
+                        nqp::while(   # doesn't sink
+                          nqp::islt_i($!i = nqp::add_i($!i,1),$elems),
+                          $target.push(
+                            nqp::ifnull(
+                              nqp::atpos($!reified,$!i),
+                              nqp::p6bindattrinvres(
+                                (my \v := nqp::p6scalarfromdesc($!descriptor)),
+                                Scalar,
+                                '$!whence',
+                                -> { nqp::bindpos($!reified,$!i,v) }
+                              )
+                            )
+                          )
+                        );
+                        IterationEnd
+                    }
+                }
+
+                method is-lazy() { $!todo.DEFINITE && $!todo.is-lazy }
+            }.new(self)
+        }
+
+        # everything we need is already there
+        elsif nqp::getattr(self,List,'$!reified').DEFINITE {
+            class :: does Iterator {
+                has int $!i;
+                has $!reified;
+                has $!descriptor;
+
+                method !SET-SELF(\array) {
+                    $!i           = -1;
+                    $!reified    := nqp::getattr(array, List,  '$!reified');
+                    $!descriptor := nqp::getattr(array, Array, '$!descriptor');
+                    self
+                }
+                method new(\array) { nqp::create(self)!SET-SELF(array) }
+
+                method pull-one() is raw {
+                    nqp::ifnull(
+                      nqp::atpos($!reified,$!i = nqp::add_i($!i,1)),
+                      nqp::islt_i($!i,nqp::elems($!reified)) # found a hole
+                        ?? nqp::p6bindattrinvres(
+                             (my \v := nqp::p6scalarfromdesc($!descriptor)),
+                             Scalar,
+                             '$!whence',
+                             -> { nqp::bindpos($!reified,$!i,v) }
+                           )
+                        !! IterationEnd
+                    )
+                }
+
+                method push-all($target) {
+                    my int $elems = nqp::elems($!reified);
+                    nqp::while(   # doesn't sink
+                      nqp::islt_i($!i = nqp::add_i($!i,1),$elems),
+                      $target.push(nqp::ifnull(
+                        nqp::atpos($!reified,$!i),
+                        nqp::p6bindattrinvres(
+                          (my \v := nqp::p6scalarfromdesc($!descriptor)),
+                          Scalar,
+                          '$!whence',
+                          -> { nqp::bindpos($!reified,$!i,v) }
+                        )
+                      ))
+                    );
+                    IterationEnd
+                }
+            }.new(self)
+        }
+
+        # nothing now or in the future to iterate over
+        else {
+            Rakudo::Internals.EmptyIterator
+        }
+    }
     method from-iterator(Array:U: Iterator $iter) {
         my \result := nqp::create(self);
         my \buffer := nqp::create(IterationBuffer);
@@ -56,36 +194,13 @@ my class Array { # declared in BOOTSTRAP
         result
     }
 
-    my constant \SHAPE-STORAGE-ROOT := do {
-        my Mu $root := nqp::newtype(nqp::knowhow(), 'Uninstantiable');
-        nqp::setparameterizer($root, -> $, $key {
-            my Mu $args := nqp::p6argvmarray();
-            my $dim_type := nqp::newtype(nqp::knowhow(), 'MultiDimArray');
-            nqp::composetype($dim_type, nqp::hash('array',
-                nqp::hash('dimensions', $key.elems)));
-            nqp::settypehll($dim_type, 'perl6');
-            $dim_type
-        });
-        nqp::settypehll($root, 'perl6');
-        $root
-    }
     sub allocate-shaped-storage(\arr, @dims) {
-        my $key := nqp::list();
-        my $dims := nqp::list_i();
-        for @dims {
-            if nqp::istype($_, Whatever) {
-                X::NYI.new(feature => 'Jagged array shapes');
-            }
-            nqp::push($key, Mu);
-            nqp::push_i($dims, $_.Int);
-        }
-        my $storage := nqp::create(nqp::parameterizetype(SHAPE-STORAGE-ROOT, $key));
-        nqp::setdimensions($storage, $dims);
-        nqp::bindattr(arr, List, '$!reified', $storage);
+        nqp::bindattr(arr, List, '$!reified',
+            Rakudo::Internals.SHAPED-ARRAY-STORAGE(@dims, nqp::knowhow(), Mu));
         arr
     }
 
-    my role ShapedArray[::TValue] does Positional[TValue] {
+    my role ShapedArray[::TValue] does Positional[TValue] does Rakudo::Internals::ShapedArrayCommon {
         has $.shape;
 
         proto method AT-POS(|) is raw {*}
@@ -98,10 +213,9 @@ my class Array { # declared in BOOTSTRAP
             my int $numind  = @indices.elems;
             if $numind >= $numdims {
                 my $idxs := nqp::list_i();
-                while $numdims > 0 {
-                    nqp::push_i($idxs, @indices.shift);
-                    $numdims = $numdims - 1;
-                }
+                nqp::push_i($idxs, @indices.shift)
+                  while nqp::isge_i(--$numdims,0);
+
                 my \elem = nqp::ifnull(
                     nqp::atposnd($storage, $idxs),
                     nqp::p6bindattrinvres(
@@ -128,10 +242,8 @@ my class Array { # declared in BOOTSTRAP
             if $numind == $numdims {
                 # Dimension counts match, so fast-path it
                 my $idxs := nqp::list_i();
-                while $numdims > 0 {
-                    nqp::push_i($idxs, @indices.shift);
-                    $numdims = $numdims - 1;
-                }
+                nqp::push_i($idxs, @indices.shift)
+                  while nqp::isge_i(--$numdims,0);
                 nqp::ifnull(
                     nqp::atposnd($storage, $idxs),
                     nqp::bindposnd($storage, $idxs,
@@ -158,30 +270,31 @@ my class Array { # declared in BOOTSTRAP
         }
         multi method EXISTS-POS(**@indices) {
             my Mu $storage := nqp::getattr(self, List, '$!reified');
+            my $dims       := nqp::dimensions($storage);
             my int $numdims = nqp::numdimensions($storage);
             my int $numind  = @indices.elems;
-            my $dims := nqp::dimensions($storage);
+            my int $i = -1;
+
             if $numind >= $numdims {
                 my $idxs := nqp::list_i();
-                loop (my int $i = 0; $i < $numind; $i = $i + 1) {
-                    my int $idx = @indices.shift;
-                    return False if $idx >= nqp::atpos_i($dims, $i);
-                    nqp::push_i($idxs, $idx);
-                }
-                if nqp::isnull(nqp::atposnd($storage, $idxs)) {
-                    False
-                }
-                elsif @indices {
-                    nqp::atposnd($storage, $idxs).EXISTS-POS(|@indices)
-                }
-                else {
-                    True
-                }
+                my int $idx;
+
+                ($idx = @indices.shift) >= nqp::atpos_i($dims,$i)
+                  ?? return False
+                  !! nqp::push_i($idxs, $idx)
+                  while nqp::islt_i(++$i,$numind);
+
+                nqp::isnull(nqp::atposnd($storage, $idxs))
+                  ?? False
+                  !! @indices
+                     ?? nqp::atposnd($storage, $idxs).EXISTS-POS(|@indices)
+                     !! True
             }
             else {
-                loop (my int $i = 0; $i < $numind; $i = $i + 1) {
-                    return False if @indices[$i] >= nqp::atpos_i($dims, $i);
-                }
+                return False
+                  if @indices[$i] >= nqp::atpos_i($dims,$i)
+                  while nqp::islt_i(++$i,$numind);
+
                 True
             }
         }
@@ -196,10 +309,9 @@ my class Array { # declared in BOOTSTRAP
             my int $numind  = @indices.elems;
             if $numind >= $numdims {
                 my $idxs := nqp::list_i();
-                while $numdims > 0 {
-                    nqp::push_i($idxs, @indices.shift);
-                    $numdims = $numdims - 1;
-                }
+                nqp::push_i($idxs, @indices.shift)
+                  while nqp::isge_i(--$numdims,0);
+
                 my \value = nqp::ifnull(nqp::atposnd($storage, $idxs), Nil);
                 if @indices {
                     value.DELETE-POS(|@indices)
@@ -223,7 +335,7 @@ my class Array { # declared in BOOTSTRAP
         multi method BIND-POS(Array:U: |c) is raw {
             self.Any::BIND-POS(|c)
         }
-        multi method BIND-POS(**@indices is raw) is raw {
+        multi method BIND-POS(Array:D: **@indices is raw) is raw {
             my Mu $storage := nqp::getattr(self, List, '$!reified');
             my int $numdims = nqp::numdimensions($storage);
             my int $numind  = @indices.elems - 1;
@@ -232,11 +344,9 @@ my class Array { # declared in BOOTSTRAP
                 # At least enough indices that binding will work out or we can
                 # pass the bind target on down the chain.
                 my $idxs := nqp::list_i();
-                my int $i = 0;
-                while $i < $numdims {
-                    nqp::push_i($idxs, @indices.AT-POS($i));
-                    $i = $i + 1;
-                }
+                my int $i = -1;
+                nqp::push_i($idxs, @indices.AT-POS($i))
+                  while nqp::islt_i(++$i,$numdims);
                 $numind == $numdims
                     ?? nqp::bindposnd($storage, $idxs, value)
                     !! nqp::atposnd($storage, $idxs).BIND-POS(|@indices[$numdims..*])
@@ -254,16 +364,15 @@ my class Array { # declared in BOOTSTRAP
         proto method STORE(|) { * }
         multi method STORE(::?CLASS:D: Iterable:D \in) {
             allocate-shaped-storage(self, self.shape);
-            if nqp::istype(in, ShapedArray) {
-                if self.shape eqv in.shape {
+            my \in-shape = nqp::can(in, 'shape') ?? in.shape !! Nil;
+            if in-shape && !nqp::istype(in-shape.AT-POS(0), Whatever) {
+                if self.shape eqv in-shape {
                     # Can do a VM-supported memcpy-like thing in the future
-                    for self.keys {
-                        self.ASSIGN-POS(|$_, in.AT-POS(|$_))
-                    }
+                    self.ASSIGN-POS(|$_, in.AT-POS(|$_)) for self.keys;
                 }
                 else {
                     X::Assignment::ArrayShapeMismatch.new(
-                        source-shape => in.shape,
+                        source-shape => in-shape,
                         target-shape => self.shape
                     ).throw
                 }
@@ -275,73 +384,17 @@ my class Array { # declared in BOOTSTRAP
         multi method STORE(::?CLASS:D: Mu \item) {
             self.STORE((item,))
         }
-        method !STORE-PATH(@path, @rest, \in) {
-            my int $cur-pos = 0;
-            if @rest.elems == 1 {
-                for in -> \item {
-                    self.ASSIGN-POS(|@path, $cur-pos, item);
-                    $cur-pos = $cur-pos + 1;
-                }
-            }
-            else {
-                my @nextrest = @rest[1..*];
-                for in -> \item {
-                    my @nextpath = flat @path, $cur-pos;
-                    if nqp::istype(item, Iterable) && nqp::isconcrete(item) {
-                        self!STORE-PATH(@nextpath, @nextrest, item)
-                    }
-                    else {
-                        X::Assignment::ToShaped.new(shape => self.shape).throw;
-                    }
-                    $cur-pos = $cur-pos + 1;
-                }
-            }
-        }
-
-        proto method push(|c) {
-            self.DEFINITE
-                ?? X::IllegalOnFixedDimensionArray.new(operation => 'push').throw
-                !! self.Any::push(|c)
-        }
-        proto method append(|c) {
-            self.DEFINITE
-                ?? X::IllegalOnFixedDimensionArray.new(operation => 'append').throw
-                !! self.Any::append(|c)
-        }
-
-        multi method pop(::?CLASS:D:) {
-            X::IllegalOnFixedDimensionArray.new(operation => 'pop').throw
-        }
-
-        multi method shift(::?CLASS:D:) {
-            X::IllegalOnFixedDimensionArray.new(operation => 'shift').throw
-        }
-
-        proto method unshift(|c) {
-            self.DEFINITE
-                ?? X::IllegalOnFixedDimensionArray.new(operation => 'unshift').throw
-                !! self.Any::unshift(|c)
-        }
-        proto method prepend(|c) {
-            self.DEFINITE
-                ?? X::IllegalOnFixedDimensionArray.new(operation => 'prepend').throw
-                !! self.Any::prepend(|c)
-        }
-
-        multi method splice(::?CLASS:D: *@) {
-            X::IllegalOnFixedDimensionArray.new(operation => 'splice').throw
-        }
-
-        multi method plan(::?CLASS:D: *@) {
-            X::IllegalOnFixedDimensionArray.new(operation => 'plan').throw
-        }
 
         method reverse(::?CLASS:D:) {
-            X::IllegalOnFixedDimensionArray.new(operation => 'reverse').throw
+            self.shape.elems == 1
+                ?? self.new(:shape(self.shape), self.List.reverse())
+                !! X::IllegalOnFixedDimensionArray.new(operation => 'reverse').throw
         }
 
-        method rotate(::?CLASS:D: Cool) {
-            X::IllegalOnFixedDimensionArray.new(operation => 'rotate').throw
+        method rotate(::?CLASS:D: Cool \n) {
+            self.shape.elems == 1
+                ?? self.new(:shape(self.shape), self.List.rotate(n))
+                !! X::IllegalOnFixedDimensionArray.new(operation => 'rotate').throw
         }
 
         # A shaped array isn't lazy, we these methods don't need to go looking
@@ -351,97 +404,6 @@ my class Array { # declared in BOOTSTRAP
         }
         method eager() { self }
         method is-lazy() { False }
-
-        multi method keys(::?CLASS:D:) {
-            my @shape := self.shape;
-            @shape.elems == 1
-                ?? Seq.new((^@shape[0]).iterator)
-                !! cross(@shape.map(^*).list)
-        }
-        multi method values(::?CLASS:D:) {
-            self.keys.map({ self.AT-POS(|$_) })
-        }
-        multi method kv(::?CLASS:D:) {
-            self.keys.map({ slip($_, self.AT-POS(|$_)) })
-        }
-        multi method pairs(::?CLASS:D:) {
-            self.keys.map({ $_ => self.AT-POS(|$_) })
-        }
-        multi method antipairs(::?CLASS:D:) {
-            self.keys.map({ self.AT-POS(|$_) => $_ })
-        }
-        multi method invert(::?CLASS:D:) {
-            self.keys.map({ nqp::decont(self.AT-POS(|$_)) »=>» $_ }).flat
-        }
-
-        method iterator(List:D:) {
-            # This can be fairly heavily optimized in various ways later
-            self.values.iterator
-        }
-
-        # These work on the flat view
-        method roll(|c) {
-            self.flat.roll(|c)
-        }
-        method pick(|c) {
-            self.flat.pick(|c)
-        }
-        method permutations(|c) {
-            self.flat.permutations(|c)
-        }
-        method combinations(|c) {
-            self.flat.combinations(|c)
-        }
-        method rotor(|c) {
-            self.flat.rotor(|c)
-        }
-        method join(|c) {
-            self.flat.join(|c)
-        }
-
-        multi method gist(::?CLASS:D:) {
-            if not %*gistseen<TOP> { my %*gistseen = :TOP ; return self.gist }
-            if %*gistseen{self.WHICH} { %*gistseen{self.WHICH} = 2; return "Array_{self.WHERE}" }
-            %*gistseen{self.WHICH} = 1;
-            my $result = self!gist([], self.shape);
-            $result = "(\\Array_{self.WHERE} = $result)" if %*gistseen{self.WHICH}:delete == 2;
-            $result;
-        }
-        method !gist(@path, @dims) {
-            if @dims.elems == 1 {
-                 '[' ~ (^@dims[0]).map({ self.AT-POS(|@path, $_).gist }).join(' ') ~ ']';
-            }
-            else {
-                my @nextdims = @dims[1..*];
-                '[' ~ (^@dims[0]).map({ self!gist((flat @path, $_), @nextdims) }).join(' ') ~ ']';
-            }
-        }
-
-        multi method perl(::?CLASS:D \SELF:) {
-            if not %*perlseen<TOP> { my %*perlseen = :TOP ; return SELF.perl }
-            if %*perlseen{self.WHICH} { %*perlseen{self.WHICH} = 2; return "Array_{self.WHERE}" }
-            %*perlseen{self.WHICH} = 1;
-            my $result = 'Array.new(:shape' ~ nqp::decont(self.shape).perl ~ ', ' ~
-                self!perl([], self.shape) ~ ')';
-            $result ~= '.item' if nqp::iscont(SELF);
-            $result = "(my \\Array_{self.WHERE} = $result)" if %*perlseen{self.WHICH}:delete == 2;
-            $result;
-        }
-        method !perl(@path, @dims) {
-            if @dims.elems == 1 {
-                 '[' ~
-                    (^@dims[0]).map({ nqp::decont(self.AT-POS(|@path, $_)).perl }).join(', ') ~
-                    ',' x (@dims[0] == 1 && nqp::istype(self.AT-POS(|@path, 0), Iterable)) ~
-                 ']'
-            }
-            else {
-                my @nextdims = @dims[1..*];
-                '[' x (@path.elems > 0) ~
-                    (^@dims[0]).map({ self!perl((flat @path, $_), @nextdims) }).join(', ') ~
-                    ',' x (@dims[0] == 1) ~
-                ']' x (@path.elems > 0)
-            }
-        }
     }
 
     proto method new(|) { * }
@@ -459,8 +421,9 @@ my class Array { # declared in BOOTSTRAP
             my \list-shape = nqp::istype(shape, List) ?? shape !! shape.list;
             allocate-shaped-storage(arr, list-shape);
             arr does ShapedArray[Mu];
+            arr.^set_name('Array');
             nqp::bindattr(arr, arr.WHAT, '$!shape', list-shape);
-            die "Creating shaped array with initial values NYI" if values;
+            arr.STORE(values) if values;
         }
         else {
             arr.STORE(values);
@@ -468,26 +431,27 @@ my class Array { # declared in BOOTSTRAP
         arr
     }
 
-    method !ensure-allocated() {
+    method !ensure-allocated(--> Nil) {
         nqp::bindattr(self, List, '$!reified', nqp::create(IterationBuffer))
             unless nqp::getattr(self, List, '$!reified').DEFINITE;
     }
 
     method is-lazy() {
         my $todo := nqp::getattr(self, List, '$!todo');
-        if $todo.DEFINITE {
-            $todo.reify-until-lazy();
-            if $todo.fully-reified {
-                nqp::bindattr(self, List, '$!todo', Mu);
-                False;
-            }
-            else {
-                True;
-            }
-        }
-        else {
-            False
-        }
+        nqp::if(
+          $todo.DEFINITE,
+          nqp::stmts(
+            $todo.reify-until-lazy,
+            nqp::if(
+              $todo.fully-reified,
+              nqp::stmts(
+                nqp::bindattr(self,List,'$!todo',Mu),
+                False
+              ),
+              True
+            )
+          )
+        )
     }
 
     proto method STORE(|) { * }
@@ -569,9 +533,8 @@ my class Array { # declared in BOOTSTRAP
             if nqp::islt_i($ipos, 0);
         self!ensure-allocated();
         my $todo := nqp::getattr(self, List, '$!todo');
-        if $todo.DEFINITE {
-            $todo.reify-at-least($ipos + 1);
-        }
+        $todo.reify-at-least($ipos + 1) if $todo.DEFINITE;
+
         my Mu \reified := nqp::getattr(self, List, '$!reified');
         $ipos >= nqp::elems(reified) || nqp::isnull(my \value = nqp::atpos(reified, $ipos))
             ?? nqp::p6bindattrinvres(
@@ -612,23 +575,22 @@ my class Array { # declared in BOOTSTRAP
             if nqp::islt_i($ipos, 0);
         self!ensure-allocated();
         my $todo := nqp::getattr(self, List, '$!todo');
-        if $todo.DEFINITE {
-            $todo.reify-at-least($ipos + 1);
-        }
+        $todo.reify-at-least($ipos + 1) if $todo.DEFINITE;
+
         my Mu \reified := nqp::getattr(self, List, '$!reified');
         $ipos >= nqp::elems(reified) || nqp::isnull(my \value = nqp::atpos(reified, $ipos))
             ?? (nqp::bindpos(reified, $ipos, nqp::p6scalarfromdesc($!descriptor)) = assignee)
             !! (nqp::atpos(reified, $ipos) = assignee)
     }
 
-    multi method BIND-POS(Int $pos, Mu \bindval) is raw {
+    multi method BIND-POS(Array:D: Int $pos, Mu \bindval) is raw {
         self!ensure-allocated();
         my int $ipos = $pos;
         my $todo := nqp::getattr(self, List, '$!todo');
         $todo.reify-at-least($ipos + 1) if $todo.DEFINITE;
         nqp::bindpos(nqp::getattr(self, List, '$!reified'), $ipos, bindval);
     }
-    multi method BIND-POS(int $pos, Mu \bindval) is raw {
+    multi method BIND-POS(Array:D: int $pos, Mu \bindval) is raw {
         self!ensure-allocated();
         my $todo := nqp::getattr(self, List, '$!todo');
         $todo.reify-at-least($pos + 1) if $todo.DEFINITE;
@@ -650,7 +612,7 @@ my class Array { # declared in BOOTSTRAP
             my int $pos = pos;
             nqp::pop($items);
             nqp::pop($items)
-              while ($pos = $pos - 1) >= 0
+              while nqp::isge_i(--$pos,0)
                 && nqp::isnull(nqp::atpos($items,$pos));
         }
         $value;
@@ -659,8 +621,9 @@ my class Array { # declared in BOOTSTRAP
     # MUST have a separate Slip variant to have it slip
     multi method push(Array:D: Slip \value) {
         self!ensure-allocated();
-        fail X::Cannot::Lazy.new(action => 'push to') if self.is-lazy;
-        self!append-list(value);
+        self.is-lazy
+          ?? Failure.new(X::Cannot::Lazy.new(action => 'push to'))
+          !! self!append-list(value)
     }
     multi method push(Array:D: \value) {
         self!ensure-allocated();
@@ -674,14 +637,17 @@ my class Array { # declared in BOOTSTRAP
     }
     multi method push(Array:D: **@values is raw) {
         self!ensure-allocated();
-        fail X::Cannot::Lazy.new(action => 'push to') if self.is-lazy;
-        self!append-list(@values)
+        self.is-lazy
+          ?? Failure.new(X::Cannot::Lazy.new(action => 'push to'))
+          !! self!append-list(@values)
     }
 
     multi method append(Array:D: \value) {
         self!ensure-allocated();
-        fail X::Cannot::Lazy.new(action => 'append to') if self.is-lazy;
-        if nqp::iscont(value) || nqp::not_i(nqp::istype(value, Iterable)) {
+        if self.is-lazy {
+            Failure.new(X::Cannot::Lazy.new(action => 'append to'))
+        }
+        elsif nqp::iscont(value) || nqp::not_i(nqp::istype(value, Iterable)) {
             nqp::push(
                 nqp::getattr(self, List, '$!reified'),
                 nqp::assign(nqp::p6scalarfromdesc($!descriptor), value)
@@ -694,18 +660,18 @@ my class Array { # declared in BOOTSTRAP
     }
     multi method append(Array:D: **@values is raw) {
         self!ensure-allocated();
-        fail X::Cannot::Lazy.new(action => 'append to') if self.is-lazy;
-        self!append-list(@values)
+        self.is-lazy
+          ?? Failure.new(X::Cannot::Lazy.new(action => 'append to'))
+          !! self!append-list(@values)
     }
     method !append-list(@values) {
         my \values-iter = @values.iterator;
         my \reified := nqp::getattr(self, List, '$!reified');
         my \target := ArrayReificationTarget.new(reified,
             nqp::decont($!descriptor));
-        unless values-iter.push-until-lazy(target) =:= IterationEnd {
-            fail X::Cannot::Lazy.new(:action<push>, :what(self.^name));
-        }
-        self
+        values-iter.push-until-lazy(target) =:= IterationEnd
+          ?? self
+          !! Failure.new(X::Cannot::Lazy.new(:action<push>, :what(self.^name)))
     }
 
     multi method unshift(Array:D: Slip \value) {
@@ -762,8 +728,8 @@ my class Array { # declared in BOOTSTRAP
 
         my $reified := nqp::getattr(self, List, '$!reified');
         nqp::elems($reified)
-            ?? nqp::pop($reified)
-            !! fail X::Cannot::Empty.new(:action<pop>, :what(self.^name));
+          ?? nqp::pop($reified)
+          !! Failure.new(X::Cannot::Empty.new(:action<pop>,:what(self.^name)))
     }
 
     method shift(Array:D:) is raw is nodal {
@@ -771,9 +737,19 @@ my class Array { # declared in BOOTSTRAP
         self!ensure-allocated();
         my $todo := nqp::getattr(self, List, '$!todo');
         my $reified := nqp::getattr(self, List, '$!reified');
-        nqp::existspos($reified, 0) || $todo.DEFINITE && $todo.reify-at-least(1)
-            ?? nqp::shift($reified)
-            !! fail X::Cannot::Empty.new(:action<shift>, :what(self.^name));
+        nqp::if(
+          (nqp::existspos($reified,0)
+            || $todo.DEFINITE && $todo.reify-at-least(1)),
+          nqp::shift($reified),
+          nqp::if(
+            nqp::elems($reified),  # is it actually just sparse?
+            nqp::stmts(
+              nqp::shift($reified),
+              Nil
+            ),
+            Failure.new(X::Cannot::Empty.new(:action<shift>,:what(self.^name)))
+          )
+        )
     }
 
     proto method splice(|) is nodal { * }
@@ -831,18 +807,14 @@ my class Array { # declared in BOOTSTRAP
         # need to enforce type checking
         my $expected := self.of;
         unless self.of =:= Mu {
-            my int $i = 0;
+            my int $i = -1;
             my int $n = nqp::elems(splice-buffer);
-            while $i < $n {
-                unless nqp::istype(nqp::atpos(splice-buffer, $i), $expected) {
-                    X::TypeCheck::Splice.new(
-                        :action<splice>,
-                        :got($_.WHAT),
-                        :$expected,
-                    ).fail;
-                }
-                $i = $i + 1;
-            }
+            fail X::TypeCheck::Splice.new(
+              :action<splice>,
+              :got($_.WHAT),
+              :$expected,
+            ) unless nqp::istype(nqp::atpos(splice-buffer,$i),$expected)
+              while nqp::islt_i(++$i,$n);
         }
 
         $todo.reify-at-least($o + $s) if $lazy;
@@ -853,7 +825,7 @@ my class Array { # declared in BOOTSTRAP
         }
         else {
             my @ret := $expected =:= Mu ?? Array.new !! Array[$expected].new;
-            @ret = self[$o..($o + $s - 1)] if $s;
+            @ret = self[lazy $o..($o + $s - 1)] if $s;
             nqp::splice(nqp::getattr(self, List, '$!reified'),
                 splice-buffer, $o, $s);
             @ret;
@@ -862,38 +834,29 @@ my class Array { # declared in BOOTSTRAP
 
     # introspection
     method name() {
-        my $d := $!descriptor;
-        nqp::isnull($d) ?? Nil !! $d.name()
+        nqp::isnull($!descriptor) ?? Nil !! $!descriptor.name
     }
     method of() {
-        my $d := $!descriptor;
-        nqp::isnull($d) ?? Mu !! $d.of;
+        nqp::isnull($!descriptor) ?? Mu !! $!descriptor.of
     }
     method default() {
-        my $d := $!descriptor;
-        nqp::isnull($d) ?? Any !! $d.default;
+        nqp::isnull($!descriptor) ?? Any !! $!descriptor.default
     }
     method dynamic() {
-        my $d := $!descriptor;
-        nqp::isnull($d) ?? Nil !! so $d.dynamic;
+        nqp::isnull($!descriptor) ?? Nil !! so $!descriptor.dynamic
     }
     multi method perl(Array:D \SELF:) {
-        if not %*perlseen<TOP> { my %*perlseen = :TOP ; return SELF.perl }
-        if %*perlseen{self.WHICH} { %*perlseen{self.WHICH} = 2; return "Array_{self.WHERE}" }
-        %*perlseen{self.WHICH} = 1;
-        my $result = '$' x nqp::iscont(SELF) ~
-        '[' ~ self.map({nqp::decont($_).perl}).join(', ') ~ ',' x (self.elems == 1 && nqp::istype(self[0],Iterable)) ~ ']';
-        $result = "(my \\Array_{self.WHERE} = $result)" if %*perlseen{self.WHICH}:delete == 2;
-        $result;
+        SELF.perlseen('Array', {
+             '$' x nqp::iscont(SELF)  # self is always deconted
+             ~ '['
+             ~ self.map({nqp::decont($_).perl}).join(', ')
+             ~ ',' x (self.elems == 1 && nqp::istype(self.AT-POS(0),Iterable))
+             ~ ']'
+        })
     }
 
     multi method gist(Array:D:) {
-        if not %*gistseen<TOP> { my %*gistseen = :TOP ; return self.gist }
-        if %*gistseen{self.WHICH} { %*gistseen{self.WHICH} = 2; return "Array_{self.WHERE}" }
-        %*gistseen{self.WHICH} = 1;
-        my $result = '[' ~ self.map({.gist}).join(' ') ~ ']';
-        $result = "(\\Array_{self.WHERE} = $result)" if %*gistseen{self.WHICH}:delete == 2;
-        $result;
+        self.gistseen('Array', { '[' ~ self.map({.gist}).join(' ') ~ ']' } )
     }
 
     multi method WHICH(Array:D:) {
@@ -907,25 +870,44 @@ my class Array { # declared in BOOTSTRAP
     }
 
     my role TypedArray[::TValue] does Positional[TValue] {
-        method new(**@values is raw) {
+        proto method new(|) { * }
+        multi method new(**@values is raw, :$shape) {
+            self!new-internal(@values, $shape);
+        }
+        multi method new(\values, :$shape) {
+            self!new-internal(values, $shape);
+        }
+
+        method !new-internal(\values, \shape) {
             my \arr = nqp::create(self);
             nqp::bindattr(
-                arr,
-                Array,
-                '$!descriptor',
-                Perl6::Metamodel::ContainerDescriptor.new(:of(TValue), :rw(1))
+              arr,
+              Array,
+              '$!descriptor',
+              Perl6::Metamodel::ContainerDescriptor.new(
+                :of(TValue), :rw(1), :default(TValue))
             );
-            arr.STORE(@values);
+            if shape.DEFINITE {
+                my \list-shape = nqp::istype(shape, List) ?? shape !! shape.list;
+                allocate-shaped-storage(arr, list-shape);
+                arr does ShapedArray[Mu];
+                arr.^set_name('Array');
+                nqp::bindattr(arr, arr.WHAT, '$!shape', list-shape);
+                arr.STORE(values) if values;
+            } else {
+                arr.STORE(values);
+            }
             arr
         }
+
         proto method BIND-POS(|) { * }
-        multi method BIND-POS(Int $pos, TValue \bindval) is raw {
+        multi method BIND-POS(Array:D: Int $pos, TValue \bindval) is raw {
             my int $ipos = $pos;
             my $todo := nqp::getattr(self, List, '$!todo');
             $todo.reify-at-least($ipos + 1) if $todo.DEFINITE;
             nqp::bindpos(nqp::getattr(self, List, '$!reified'), $ipos, bindval)
         }
-        multi method BIND-POS(int $pos, TValue \bindval) is raw {
+        multi method BIND-POS(Array:D: int $pos, TValue \bindval) is raw {
             my $todo := nqp::getattr(self, List, '$!todo');
             $todo.reify-at-least($pos + 1) if $todo.DEFINITE;
             nqp::bindpos(nqp::getattr(self, List, '$!reified'), $pos, bindval)
