@@ -1,6 +1,7 @@
 my class Exception { ... }
 
 my class Backtrace { ... }
+my class CompUnit::RepositoryRegistry is repr('Uninstantiable') { ... }
 
 my $RAKUDO-VERBOSE-STACKFRAME;
 
@@ -22,7 +23,7 @@ my class Backtrace::Frame {
     multi method Str(Backtrace::Frame:D:) {
         my $s = self.subtype;
         $s ~= ' ' if $s.chars;
-        my $text = "  in {$s}$.subname at {$.file}:$.line\n";
+        my $text = "  in {$s}$.subname at {$.file} line $.line\n";
 
         if $RAKUDO-VERBOSE-STACKFRAME -> $extra {
             my $io = $!file.IO;
@@ -50,7 +51,7 @@ my class Backtrace {
     has Mu $!frames;
     has Int $!bt-next;   # next bt index to vivify
 
-    submethod BUILD(:$!bt, :$!bt-next) { $!frames := nqp::list }
+    submethod BUILD(:$!bt, :$!bt-next --> Nil) { $!frames := nqp::list }
 
     multi method new(Mu $e, Int $offset = 0) {
         $e.^name eq 'BOOTException'
@@ -94,6 +95,10 @@ my class Backtrace {
             my $file := $annotations<file>;
             next unless $file;
 
+            if CompUnit::RepositoryRegistry.file-for-spec($file) -> $path {
+                $file := $path.abspath;
+            }
+
             # now *that's* an evil hack
             next if $file.ends-with('BOOTSTRAP.nqp')
                  || $file.ends-with('QRegex.nqp')
@@ -158,7 +163,7 @@ my class Backtrace {
 
             return $idx - 1;
         }
-        Int;
+        Nil;
     }
 
     method outer-caller-idx(Backtrace:D: Int $startidx) {
@@ -203,7 +208,7 @@ my class Backtrace {
                     @frames.push: $prev;
                 } else {
                     my @outer_callers := self.outer-caller-idx($i);
-                    my ($target_idx) = @outer_callers.keys.grep({self.AT-POS($i).code.^isa(Routine)});
+                    my $target_idx = @outer_callers.keys.grep({self.AT-POS($i).code.^isa(Routine)})[0];
                     $target_idx    ||= @outer_callers[0] || $i;
                     my $current = self.AT-POS($target_idx);
                     @frames.append: $current.clone(line => $prev.line);
