@@ -53,7 +53,7 @@ my role Baggy does QuantHash {
     }
 
 #--- interface methods
-    method BUILD(Baggy:D: Mu \elems) {
+    method !SET-SELF(Baggy:D: Mu \elems) {
         %!elems := elems;
 
         if self.^name.chars == 3 { # shoddy heuristic for Bag/Mix
@@ -114,7 +114,7 @@ my role Baggy does QuantHash {
                 nqp::bindkey($elems,$which,self!PAIR($_,1));
             }
         }
-        nqp::create(self).BUILD($elems)
+        nqp::create(self)!SET-SELF($elems)
     }
     method new-from-pairs(*@pairs) {
         my $elems := nqp::hash();
@@ -146,32 +146,22 @@ my role Baggy does QuantHash {
             }
         }
         self!SANITY($elems) if $seen-pair;
-        nqp::create(self).BUILD($elems)
+        nqp::create(self)!SET-SELF($elems)
     }
 
 #--- iterator methods
     multi method pairs(Baggy:D:) {
         Seq.new(class :: does Rakudo::Internals::MappyIterator {
             method pull-one() {
-                $!hash-iter
-                  ?? nqp::iterval(nqp::shift($!hash-iter))
+                $!iter
+                  ?? nqp::iterval(nqp::shift($!iter))
                   !! IterationEnd
             }
-            method push-exactly($target, int $n) {
-                my int $done;
-                my $no-sink;
-                while $done < $n {
-                    return IterationEnd unless $!hash-iter;
-                    $no-sink :=
-                      $target.push(nqp::iterval(nqp::shift($!hash-iter)));
-                    $done = $done + 1;
-                }
-                $done
-            }
             method push-all($target) {
-                my $no-sink;
-                $no-sink := $target.push(nqp::iterval(nqp::shift($!hash-iter)))
-                  while $!hash-iter;
+                nqp::while(  # doesn't sink
+                  $!iter,
+                  $target.push(nqp::iterval(nqp::shift($!iter)))
+                );
                 IterationEnd
             }
         }.new(%!elems))
@@ -179,26 +169,15 @@ my role Baggy does QuantHash {
     multi method keys(Baggy:D:) {
         Seq.new(class :: does Rakudo::Internals::MappyIterator {
             method pull-one() {
-                $!hash-iter
-                  ?? nqp::iterval(nqp::shift($!hash-iter)).key
+                $!iter
+                  ?? nqp::iterval(nqp::shift($!iter)).key
                   !! IterationEnd
             }
-            method push-exactly($target, int $n) {
-                my int $done;
-                my $no-sink;
-                while $done < $n {
-                    return IterationEnd unless $!hash-iter;
-                    $no-sink :=
-                      $target.push(nqp::iterval(nqp::shift($!hash-iter)).key);
-                    $done = $done + 1;
-                }
-                $done
-            }
             method push-all($target) {
-                my $no-sink;
-                $no-sink :=
-                  $target.push(nqp::iterval(nqp::shift($!hash-iter)).key)
-                    while $!hash-iter;
+                nqp::while(  # doesn't sink
+                  $!iter,
+                  $target.push(nqp::iterval(nqp::shift($!iter)).key)
+                );
                 IterationEnd
             }
         }.new(%!elems))
@@ -213,9 +192,8 @@ my role Baggy does QuantHash {
                     $!value := Mu;
                     tmp
                 }
-                elsif $!hash-iter {
-                    my \tmp =
-                      nqp::decont(nqp::iterval(nqp::shift($!hash-iter)));
+                elsif $!iter {
+                    my \tmp = nqp::decont(nqp::iterval(nqp::shift($!iter)));
                     $!value := nqp::getattr(tmp,Pair,'$!value');
                     nqp::getattr(tmp,Pair,'$!key')
                 }
@@ -223,38 +201,16 @@ my role Baggy does QuantHash {
                     IterationEnd
                 }
             }
-            method push-exactly($target, int $n) {
-                my int $done;
-                my $no-sink;
-                if $!value.DEFINITE {
-                    $no-sink := $target.push($!value);
-                    $!value := Mu;
-                    $done = $done + 1;
-                }
-                while $done < $n {
-                    return IterationEnd unless $!hash-iter;
-                    my \tmp =
-                      nqp::decont(nqp::iterval(nqp::shift($!hash-iter)));
-                    $no-sink := $target.push(nqp::getattr(tmp,Pair,'$!key'));
-                    if ($done = $done + 1) < $n {
-                        $no-sink :=
-                          $target.push(nqp::getattr(tmp,Pair,'$!value'));
-                        $done = $done + 1;
-                    }
-                    else {
-                        $!value := nqp::getattr(tmp,Pair,'$!value');
-                    }
-                }
-                $done
-            }
             method push-all($target) {
-                my $no-sink;
-                while $!hash-iter {
-                    my \tmp =
-                      nqp::decont(nqp::iterval(nqp::shift($!hash-iter)));
-                    $no-sink := $target.push(nqp::getattr(tmp,Pair,'$!key'));
-                    $no-sink := $target.push(nqp::getattr(tmp,Pair,'$!value'));
-                }
+                my $tmp;
+                nqp::while(
+                  $!iter,
+                  nqp::stmts(  # doesn't sink
+                    ($tmp := nqp::decont(nqp::iterval(nqp::shift($!iter)))),
+                    ($target.push(nqp::getattr($tmp,Pair,'$!key'))),
+                    ($target.push(nqp::getattr($tmp,Pair,'$!value')))
+                  )
+                );
                 IterationEnd
             }
         }.new(%!elems))
@@ -262,27 +218,17 @@ my role Baggy does QuantHash {
     multi method values(Baggy:D:) {
         Seq.new(class :: does Rakudo::Internals::MappyIterator {
             method pull-one() is raw {
-                $!hash-iter
+                $!iter
                     ?? nqp::getattr(nqp::decont(
-                         nqp::iterval(nqp::shift($!hash-iter))),Pair,'$!value')
+                         nqp::iterval(nqp::shift($!iter))),Pair,'$!value')
                     !! IterationEnd
             }
-            method push-exactly($target, int $n) {
-                my int $done;
-                my $no-sink;
-                while $done < $n {
-                    return IterationEnd unless $!hash-iter;
-                    $no-sink := $target.push(nqp::getattr(nqp::decont(
-                      nqp::iterval(nqp::shift($!hash-iter))),Pair,'$!value'));
-                    $done = $done + 1;
-                }
-                $done
-            }
             method push-all($target) {
-                my $no-sink;
-                $no-sink := $target.push(nqp::getattr(nqp::decont(
-                  nqp::iterval(nqp::shift($!hash-iter))),Pair,'$!value'
-                )) while $!hash-iter;
+                nqp::while(  # doesn't sink
+                  $!iter,
+                  $target.push(nqp::getattr(nqp::decont(
+                    nqp::iterval(nqp::shift($!iter))),Pair,'$!value'))
+                );
                 IterationEnd
             }
         }.new(%!elems))
@@ -290,31 +236,23 @@ my role Baggy does QuantHash {
     multi method antipairs(Baggy:D:) {
         Seq.new(class :: does Rakudo::Internals::MappyIterator {
             method pull-one() {
-                if $!hash-iter {
-                    my \tmp = nqp::iterval(nqp::shift($!hash-iter));
+                if $!iter {
+                    my \tmp = nqp::iterval(nqp::shift($!iter));
                     Pair.new(tmp.value, tmp.key)
                 }
                 else {
                     IterationEnd
                 }
             }
-            method push-exactly($target, int $n) {
-                my int $done;
-                my $no-sink;
-                while $done < $n {
-                    return IterationEnd unless $!hash-iter;
-                    my \tmp = nqp::iterval(nqp::shift($!hash-iter));
-                    $no-sink := $target.push(Pair.new(tmp.value, tmp.key));
-                    $done = $done + 1;
-                }
-                IterationEnd
-            }
             method push-all($target) {
-                my $no-sink;
-                while $!hash-iter {
-                    my \tmp = nqp::iterval(nqp::shift($!hash-iter));
-                    $no-sink := $target.push(Pair.new(tmp.value, tmp.key));
-                }
+                my $tmp;
+                nqp::while(
+                  $!iter,
+                  nqp::stmts(  # doesn't sink
+                    ($tmp := nqp::iterval(nqp::shift($!iter))),
+                    ($target.push(Pair.new($tmp.value,$tmp.key)))
+                  )
+                );
                 IterationEnd
             }
         }.new(%!elems))
@@ -330,8 +268,8 @@ my role Baggy does QuantHash {
                     $!times = $!times - 1;
                     $!key
                 }
-                elsif $!hash-iter {
-                    my \tmp = nqp::iterval(nqp::shift($!hash-iter));
+                elsif $!iter {
+                    my \tmp = nqp::iterval(nqp::shift($!iter));
                     $!key  := tmp.key;
                     $!times = tmp.value - 1;
                     $!key
@@ -341,13 +279,19 @@ my role Baggy does QuantHash {
                 }
             }
             method push-all($target) {
-                my $no-sink;
-                while $!hash-iter {
-                    my \tmp = nqp::iterval(nqp::shift($!hash-iter));
-                    $!key  := tmp.key;
-                    $!times = tmp.value + 1;
-                    $no-sink := $target.push($!key) while $!times = $!times - 1;
-                }
+                my $tmp;
+                nqp::while(
+                  $!iter,
+                  nqp::stmts(
+                    ($tmp   := nqp::iterval(nqp::shift($!iter))),
+                    ($!key  := $tmp.key),
+                    ($!times = nqp::add_i($tmp.value,1)),
+                    nqp::while(  # doesn't sink
+                      ($!times = nqp::sub_i($!times,1)),
+                      ($target.push($!key))
+                    )
+                  )
+                );
                 IterationEnd
             }
         }.new(%!elems))
@@ -469,13 +413,13 @@ my role Baggy does QuantHash {
             has int $!todo;
             has int $!keep;
 
-            method BUILD($!total, @!pairs, \keep, \todo) {
+            method !SET-SELF($!total, @!pairs, \keep, \todo) {
                 $!todo = todo;
                 $!keep = +?keep;
                 self
             }
             method new(\total,\pairs,\keep,\count) {
-                nqp::create(self).BUILD(
+                nqp::create(self)!SET-SELF(
                   total, pairs, keep, keep ?? count !! (total min count))
             }
 
@@ -501,8 +445,10 @@ my role Baggy does QuantHash {
             has Int $!total;
             has @!pairs;
 
-            method BUILD($!total, @!pairs) { self }
-            method new(\total,\pairs) { nqp::create(self).BUILD(total,pairs) }
+            method !SET-SELF($!total, @!pairs) { self }
+            method new(\total,\pairs) {
+                nqp::create(self)!SET-SELF(total,pairs)
+            }
             method is-lazy() { True }
 
             method pull-one() {
