@@ -1,14 +1,6 @@
 my role  IO         { ... }
 my class IO::Path   { ... }
 
-my class SprintfHandler {
-    method mine($x) { nqp::reprname($x) eq "P6opaque"; }
-
-    method int($x) { $x.Int }
-}
-
-my $sprintfHandlerInitialized = False;
-
 my class Cool { # declared in BOOTSTRAP
     # class Cool is Any {
 
@@ -76,10 +68,7 @@ my class Cool { # declared in BOOTSTRAP
     }
 
     method fmt($format = '%s') {
-        unless $sprintfHandlerInitialized {
-            nqp::sprintfaddargumenthandler(SprintfHandler.new);
-            $sprintfHandlerInitialized = True;
-        }
+        Rakudo::Internals.initialize-sprintf-handler;
         nqp::p6box_s(
             nqp::sprintf(nqp::unbox_s($format.Stringy), nqp::list(self))
         )
@@ -134,9 +123,7 @@ my class Cool { # declared in BOOTSTRAP
     method chr() {
         self.Int.chr;
     }
-    method chrs(Cool:D:) {
-        self>>.chr.join;
-    }
+    method chrs(Cool:D:) { chrs(self.list) }
     method ords(Cool:D:) { self.Str.ords }
 
 
@@ -166,7 +153,7 @@ my class Cool { # declared in BOOTSTRAP
         my str $str = nqp::unbox_s(self.Str);
         my int $pos =
           nqp::defined($start) ?? nqp::chars($str) min $start.Int !! 0;
-        $pos >= 0 && nqp::eqat($str, nqp::unbox_s($needle), $pos);
+        $pos >= 0 && nqp::p6bool(nqp::eqat($str, nqp::unbox_s($needle), $pos));
     }
 
     proto method contains(|) {*}
@@ -200,15 +187,17 @@ my class Cool { # declared in BOOTSTRAP
         $i < 0 ?? Nil !! nqp::box_i($i,Int);
     }
     multi method index(Cool:D: Str(Cool) $needle, Int(Cool) $pos) {
-        my int $i;
-        try {
-            $i = nqp::unbox_i($pos);
-            CATCH {
-                default {
-                    return Nil
-                }
-            }
-        }
+        fail X::OutOfRange.new(
+          :what("Position in index"),
+          :got($pos),
+          :range("0..{self.chars}"),
+        ) if nqp::isbig_I(nqp::decont($pos));
+        my int $i = nqp::unbox_i($pos);
+        fail X::OutOfRange.new(
+          :what("Position in index"),
+          :got($i),
+          :range("0..{self.chars}"),
+        ) if $i < 0;
         $i = nqp::index(
           nqp::unbox_s(self.Str),
           nqp::unbox_s($needle),
@@ -223,15 +212,17 @@ my class Cool { # declared in BOOTSTRAP
         $i < 0 ?? Nil !! nqp::box_i($i,Int);
     }
     multi method rindex(Cool:D: Str(Cool) $needle, Int(Cool) $pos) {
-        my int $i;
-        try {
-            $i = nqp::unbox_i($pos);
-            CATCH {
-                default {
-                    return Nil
-                }
-            }
-        }
+        fail X::OutOfRange.new(
+          :what("Position in rindex"),
+          :got($pos),
+          :range("0..{self.chars}"),
+        ) if nqp::isbig_I(nqp::decont($pos));
+        my int $i = nqp::unbox_i($pos);
+        fail X::OutOfRange.new(
+          :what("Position in rindex"),
+          :got($i),
+          :range("0..{self.chars}"),
+        ) if $i < 0;
         $i = nqp::rindex(
           nqp::unbox_s(self.Str),
           nqp::unbox_s($needle),
@@ -271,16 +262,18 @@ my class Cool { # declared in BOOTSTRAP
         self.Stringy.subst($matcher, $replacement, |%adverbs);
     }
 
-    proto method subst-mutate(|c) {
+    proto method subst-mutate(|) {
         $/ := nqp::getlexdyn('$/');
         {*}
     }
-    multi method subst-mutate($self is rw: |c) {
+    multi method subst-mutate(
+      Cool:D $self is rw: $matcher, $replacement, *%named
+    ) {
         $/ := nqp::getlexdyn('$/');
-        my $str = Str($self);
-        my $match = $str.subst-mutate(|c);
-        $self = $str;
-        $match;
+        my $str   = $self.Str;
+        my $match = $str.subst-mutate($matcher,$replacement,|%named);
+        $self     = $str;
+        $match
     }
 
     proto method IO(|) { * }
@@ -307,12 +300,12 @@ my class Cool { # declared in BOOTSTRAP
     proto method UInt(|) { * }
     multi method UInt()  {
         my $got := self.Int;
-        fail X::OutOfRange.new(
-          :what('Coercion to UInt'),
-          :$got,
-          :range("0..Inf")
-        ) if $got < 0;
-        $got;
+        $got < 0
+          ?? Failure.new(X::OutOfRange.new(
+               :what('Coercion to UInt'),
+               :$got,
+               :range("0..Inf")))
+          !! $got
     }
 
     method Num()  { self.Numeric.Num }
@@ -352,17 +345,21 @@ multi sub wordcase(Str:D $x) {$x.wordcase }
 multi sub wordcase(Cool $x)  {$x.Str.wordcase }
 
 sub sprintf(Cool $format, *@args) {
-    unless $sprintfHandlerInitialized {
-        nqp::sprintfaddargumenthandler(SprintfHandler.new);
-        $sprintfHandlerInitialized = True;
+    CATCH {
+        when X::Cannot::Lazy {
+            X::Cannot::Lazy.new(:action('(s)printf')).throw
+        }
+        default {
+            Rakudo::Internals.HANDLE-NQP-SPRINTF-ERRORS($_).throw
+        }
     }
-
+    Rakudo::Internals.initialize-sprintf-handler;
     @args.elems;
     nqp::p6box_s(
         nqp::sprintf(nqp::unbox_s($format.Stringy),
             nqp::clone(nqp::getattr(@args, List, '$!reified'))
         )
-    );
+    )
 }
 
 sub printf(Cool $format, *@args)          { print sprintf $format, @args }
