@@ -15,7 +15,7 @@ my class Exception {
 
     # Only valid if .backtrace has not been called yet
     method vault-backtrace(Exception:D:) {
-	nqp::isconcrete($!ex) && $!bt ?? Backtrace.new($!ex) !! ''
+        nqp::isconcrete($!ex) && $!bt ?? Backtrace.new($!ex) !! ''
     }
     method reset-backtrace(Exception:D:) {
         nqp::bindattr(self, Exception, '$!ex', Nil)
@@ -80,14 +80,39 @@ my class Exception {
     method fail(Exception:D:) {
         try self.throw;
         my $fail := Failure.new($!);
-        my Mu $return := nqp::getlexcaller('RETURN');
-        $return($fail) unless nqp::isnull($return);
-        $fail
+        nqp::throwpayloadlexcaller(nqp::const::CONTROL_RETURN, $fail);
+        CATCH { $fail.exception.throw }
     }
 
     method is-compile-time { False }
 }
 
+my class X::SecurityPolicy is Exception {}
+
+my class X::SecurityPolicy::Eval is X::SecurityPolicy {
+    has $.payload = "EVAL is a very dangerous function!!!";
+
+    my role SlurpySentry { }
+
+    method message() {
+        do {
+            # Remove spaces for die(*@msg)/fail(*@msg) forms
+            given $.payload {
+                when SlurpySentry {
+                    $_.list.join;
+                }
+                default {
+                    .Str;
+                }
+            }
+        } ~ " (use MONKEY-SEE-NO-EVAL to override,\nbut only if you're VERY sure your data contains no injection attacks)";
+    }
+    method Numeric() { $.payload.Numeric }
+    method from-slurpy (|cap) {
+        self.new(:payload(cap does SlurpySentry))
+    }
+}
+
 my class X::AdHoc is Exception {
     has $.payload = "Unexplained error";
 
@@ -110,6 +135,12 @@ my class X::AdHoc is Exception {
     }
 }
 
+my class X::NQP::NotFound is Exception {
+    has $.op;
+    method message() {
+        "Could not find nqp::$.op, did you forget 'use nqp;' ?"
+    }
+}
 my class X::Dynamic::NotFound is Exception {
     has $.name;
     method message() {
@@ -158,7 +189,10 @@ my class X::Pragma::NoArgs is Exception {
     has $.name;
     method message { "The '$.name' pragma does not take any arguments." }
 }
-
+my class X::Pragma::CannotPrecomp is Exception {
+    has $.what = 'This compilation unit';
+    method message { "$.what may not be pre-compiled" }
+}
 my class X::Pragma::CannotWhat is Exception {
     has $.what;
     has $.name;
@@ -174,7 +208,6 @@ my class X::Pragma::UnknownArg is Exception {
     has $.arg;
     method message { "Unknown argument '{$.arg.perl}' specified with the '$.name' pragma." }
 }
-
 my class X::Pragma::OnlyOne is Exception {
     has $.name;
     method message { "The '$.name' pragma only takes one argument." }
@@ -287,6 +320,9 @@ do {
                    nqp::printfh($err, "\n");
                 }
             }
+            elsif Rakudo::Internals.VERBATIM-EXCEPTION(0) {
+                nqp::printfh($err, $e.Str);
+            }
             else {
                 nqp::printfh($err, "===SORRY!===\n");
                 nqp::printfh($err, $e.Str);
@@ -312,29 +348,30 @@ do {
             nqp::printfh($err, $backtrace.first-none-setting-line);
             nqp::resume($ex)
         }
-        if $type == nqp::const::CONTROL_LAST {
-            X::ControlFlow.new(illegal => 'last', enclosing => 'loop construct', :$backtrace).throw;
+        my $label = $type +& nqp::const::CONTROL_LABELED ?? "labeled " !! "";
+        if $type +& nqp::const::CONTROL_LAST {
+            X::ControlFlow.new(illegal => "{$label}last", enclosing => 'loop construct', :$backtrace).throw;
         }
-        elsif $type == nqp::const::CONTROL_NEXT {
-            X::ControlFlow.new(illegal => 'next', enclosing => 'loop construct', :$backtrace).throw;
+        elsif $type +& nqp::const::CONTROL_NEXT {
+            X::ControlFlow.new(illegal => "{$label}next", enclosing => 'loop construct', :$backtrace).throw;
         }
-        elsif $type == nqp::const::CONTROL_REDO {
-            X::ControlFlow.new(illegal => 'redo', enclosing => 'loop construct', :$backtrace).throw;
+        elsif $type +& nqp::const::CONTROL_REDO {
+            X::ControlFlow.new(illegal => "{$label}redo", enclosing => 'loop construct', :$backtrace).throw;
         }
-        elsif $type == nqp::const::CONTROL_PROCEED {
+        elsif $type +& nqp::const::CONTROL_PROCEED {
             X::ControlFlow.new(illegal => 'proceed', enclosing => 'when clause', :$backtrace).throw;
         }
-        elsif $type == nqp::const::CONTROL_SUCCEED {
+        elsif $type +& nqp::const::CONTROL_SUCCEED {
             # XXX: should work like leave() ?
             X::ControlFlow.new(illegal => 'succeed', enclosing => 'when clause', :$backtrace).throw;
         }
-        elsif $type == nqp::const::CONTROL_TAKE {
+        elsif $type +& nqp::const::CONTROL_TAKE {
             X::ControlFlow.new(illegal => 'take', enclosing => 'gather', :$backtrace).throw;
         }
-        elsif $type == nqp::const::CONTROL_EMIT {
+        elsif $type +& nqp::const::CONTROL_EMIT {
             X::ControlFlow.new(illegal => 'emit', enclosing => 'supply or react', :$backtrace).throw;
         }
-        elsif $type == nqp::const::CONTROL_DONE {
+        elsif $type +& nqp::const::CONTROL_DONE {
             X::ControlFlow.new(illegal => 'done', enclosing => 'supply or react', :$backtrace).throw;
         }
         else {
@@ -368,6 +405,10 @@ my role X::OS is Exception {
 
 my role X::IO does X::OS { };
 
+my class X::IO::Unknown does X::IO {
+    has $.trying;
+    method message { "Unknown IO error trying '$.trying'" }
+}
 my class X::IO::Rename does X::IO {
     has $.from;
     has $.to;
@@ -487,6 +528,7 @@ my class X::IO::Chmod does X::IO {
 
 my role X::Comp is Exception {
     has $.filename;
+    has $.pos;
     has $.line;
     has $.column;
     has @.modules;
@@ -496,12 +538,8 @@ my role X::Comp is Exception {
     has @.highexpect;
     multi method gist(::?CLASS:D: :$sorry = True, :$expect = True) {
         if $.is-compile-time {
-            my $is-win := $*DISTRO.is-win;
-            my $color = %*ENV<RAKUDO_ERROR_COLOR> // !$is-win;
-            my ($red, $green, $yellow, $clear) = $color
-                ?? ("\e[31m", "\e[32m", "\e[33m", "\e[0m")
-                !! ("", "", "", "");
-            my $eject = $is-win ?? "<HERE>" !! "\x[23CF]";
+            my ($red,$clear,$green,$yellow,$eject) =
+              Rakudo::Internals.error-rcgye;
             my $r = $sorry ?? self.sorry_heading() !! "";
             $r ~= "$.message\nat $.filename():$.line";
             $r ~= "\n------> $green$.pre$yellow$eject$red$.post$clear" if defined $.pre;
@@ -514,8 +552,8 @@ my role X::Comp is Exception {
             for @.modules.reverse[1..*] {
                 my $line = nqp::p6box_i($_<line>);
                 $r ~= $_<module>.defined
-                        ?? "\n  from module $_<module> ($_<filename>:$line)"
-                        !! "\n  from $_<filename>:$line";
+                        ?? "\n  from module $_<module> ($_<filename> line $line)"
+                        !! "\n  from $_<filename> line $line";
             }
             $r;
         }
@@ -524,8 +562,7 @@ my role X::Comp is Exception {
         }
     }
     method sorry_heading() {
-        my $color = %*ENV<RAKUDO_ERROR_COLOR> // !$*DISTRO.is-win;
-        my ($red, $clear) = $color ?? ("\e[31m", "\e[0m") !! ("", "");
+        my ($red, $clear) = Rakudo::Internals.error-rcgye;
         "$red==={$clear}SORRY!$red===$clear Error while compiling $.filename\n"
     }
     method SET_FILE_LINE($file, $line) {
@@ -545,8 +582,7 @@ my class X::Comp::Group is Exception {
     multi method gist(::?CLASS:D:) {
         my $r = "";
         if $.panic || @.sorrows {
-            my $color = %*ENV<RAKUDO_ERROR_COLOR> // !$*DISTRO.is-win;
-            my ($red, $clear) = $color ?? ("\e[31m", "\e[0m") !! ("", "");
+            my ($red, $clear) = Rakudo::Internals.error-rcgye;
             $r ~= "$red==={$clear}SORRY!$red===$clear\n";
             for @.sorrows {
                 $r ~= .gist(:!sorry, :!expect) ~ "\n";
@@ -599,8 +635,8 @@ my class X::Comp::BeginTime does X::Comp {
         for @.modules.reverse[1..*] {
             my $line = nqp::p6box_i($_<line>);
             $r ~= $_<module>.defined
-                    ?? "\n  from module $_<module> ($_<filename>:$line)"
-                    !! "\n  from $_<filename>:$line";
+                    ?? "\n  from module $_<module> ($_<filename> line $line)"
+                    !! "\n  from $_<filename> line $line";
         }
         unless $!exception ~~ X::MOP {
             $r ~= "\nException details:\n" ~ $!exception.gist.indent(2);
@@ -614,6 +650,15 @@ my class X::Comp::AdHoc is X::AdHoc does X::Comp {
     method is-compile-time() { True }
 }
 
+my class X::Comp::FailGoal does X::Comp {
+    has $.dba;
+    has $.goal;
+
+    method is-compile-time() { True }
+
+    method message { "Unable to parse expression in $.dba; couldn't find final $.goal" }
+}
+
 my role X::Syntax does X::Comp { }
 my role X::Pod                 { }
 
@@ -633,6 +678,19 @@ my class X::NYI::Available is X::NYI {
         "Please install { self.available-str } for $.feature support. "
     }
 }
+my class X::NYI::BigInt is Exception {
+    has $.op;
+    has $.big;
+    has $.side = 'right';
+    method message() {
+        "Big integer $!big not yet supported on {$!side}hand side of '$!op' operator"
+    }
+}
+my class X::Experimental does X::Comp {
+    has $.feature;
+    has $.use = $!feature;
+    method message() { "Use of $.feature is experimental; please 'use experimental :$.use'" }
+}
 
 my class X::Worry is Exception { }
 my class X::Worry::P5 is X::Worry { }
@@ -1053,15 +1111,17 @@ my class X::Adverb is Exception {
         my $text = '';
         if @!unexpected.elems -> $elems {
             $text = $elems > 1
-              ?? "$elems unexpected adverbs (@!unexpected.sort())"
+              ?? "$elems unexpected adverbs (@.unexpected[])"
               !! "Unexpected adverb '@!unexpected[0]'"
         }
         if @!nogo {
             $text ~= $text ?? " and u" !! "U";
-            $text ~= "nsupported combination of adverbs (@!nogo.sort())";
+            $text ~= "nsupported combination of adverbs (@.nogo[])";
         }
         $text ~ " passed to $!what on $!source";
     }
+    method unexpected { @!unexpected.sort }
+    method nogo       { @!nogo.sort }
 }
 
 my class X::Bind is Exception {
@@ -1098,6 +1158,15 @@ my class X::Subscript::Negative is Exception {
     }
 }
 
+my class X::Invalid::Value is Exception {
+    has $.method;
+    has $.name;
+    has $.value;
+    method message {
+        "Invalid value '$.value' for :$.name on method $.method"
+    }
+}
+
 my class X::Value::Dynamic does X::Comp {
     has $.what;
     method message() { "$.what value must be known at compile time" }
@@ -1250,9 +1319,11 @@ my class X::Syntax::Missing does X::Syntax {
 my class X::Syntax::BlockGobbled does X::Syntax {
     has $.what;
     method message() {
+        my $looks_like_type = $.what ~~ /'::' | <[A..Z]><[a..z]>+/;
         $.what ~~ /^'is '/
             ?? "Trait '$.what' needs whitespace before block"
-            !! "{ $.what ?? "Function '$.what'" !! 'Expression' } needs parens to avoid gobbling block";
+            !! "{ $.what ?? "Function '$.what'" !! 'Expression' } needs parens to avoid gobbling block" ~
+                    ($looks_like_type ?? " (or perhaps it's a class that's not declared or available in this scope?)" !! "");
     };
 }
 
@@ -1285,7 +1356,7 @@ my class X::Syntax::Perl5Var does X::Syntax {
       '$`'  => '$/.prematch',
       '$\'' => '$/.postmatch',
       '$,'  => '$*OUT.output_field_separator()',
-      '$.'  => "the filehandle's .ins method",
+      '$.'  => "the .kv method on e.g. .lines",
       '$/'  => "the filehandle's .nl-in attribute",
       '$\\' => "the filehandle's .nl-out attribute",
       '$|'  => ':autoflush on open',
@@ -1306,7 +1377,7 @@ my class X::Syntax::Perl5Var does X::Syntax {
       '$^O' => '$?DISTRO.name or $*DISTRO.name',
       '$^R' => 'an explicit result variable',
       '$^S' => 'context function',
-      '$^T' => '$*BASETIME',
+      '$^T' => '$*INITTIME',
       '$^V' => '$*PERL.version or $*PERL.compiler.version',
       '$^W' => '$*WARNING',
       '$^X' => '$*EXECUTABLE-NAME',
@@ -1343,7 +1414,7 @@ my class X::Syntax::Self::WithoutObject does X::Syntax {
 }
 my class X::Syntax::VirtualCall does X::Syntax {
     has $.call;
-    method message() { "Virtual call $.call may not be used on partially constructed objects" }
+    method message() { "Virtual method call $.call may not be used on partially constructed object (maybe you mean {$.call.subst('.','!')} for direct attribute access here?)" }
 }
 my class X::Syntax::NoSelf does X::Syntax {
     has $.variable;
@@ -1379,7 +1450,7 @@ my class X::Syntax::NonAssociative does X::Syntax {
     has $.left;
     has $.right;
     method message() {
-        "Operators '$.left' and '$.right' are non-associative and require parentheses";
+        "Only identical operators may be list associative; since '$.left' and '$.right' differ, they are non-associative and you need to clarify with parentheses";
     }
 }
 
@@ -1396,7 +1467,7 @@ my class X::Syntax::CannotMeta does X::Syntax {
 my class X::Syntax::Adverb does X::Syntax {
     has $.what;
 
-    method message() { "You can't adverb " ~ ($.what // "that")  }
+    method message() { "You can't adverb " ~ $.what }
 }
 
 my class X::Syntax::Regex::Adverb does X::Syntax {
@@ -1548,6 +1619,13 @@ my class X::Attribute::Required does X::MOP {
     has $.name;
     method message() { "The attribute '$.name' is required, but you did not provide a value for it." }
 }
+my class X::Attribute::Scope::Package does X::Comp {
+    has $.scope;
+    has $.allowed;
+    has $.disallowed;
+    method message() { "Cannot use {$.scope}-scoped attribute in $.disallowed"
+        ~ ($.allowed ?? ", only $.allowed." !! ".") }
+}
 my class X::Declaration::Scope does X::Comp {
     has $.scope;
     has $.declaration;
@@ -1606,6 +1684,13 @@ my class X::Match::Bool is Exception {
     method message() { "Cannot use Bool as Matcher with '" ~ $.type ~ "'.  Did you mean to use \$_ inside a block?" }
 }
 
+my class X::LibNone does X::Comp {
+    method message { q/Must specify at least one repository.  Did you mean 'use lib "lib"' ?/ }
+}
+my class X::Package::UseLib does X::Comp {
+    has $.what;
+    method message { "Cannot 'use lib' inside a $.what" }
+}
 my class X::Package::Stubbed does X::Comp {
     has @.packages;
     # TODO: suppress display of line number
@@ -1631,22 +1716,18 @@ my class X::Str::Numeric is Exception {
     has $.pos;
     has $.reason;
     method source-indicator {
-        my $color = %*ENV<RAKUDO_ERROR_COLOR> // !$*DISTRO.is-win;
-        my $goodpart = $color ?? "\e[32m" !! "";
-        my $ejection = $color ?? "\e[33m\c[EJECT SYMBOL]" !! "\c[EJECT SYMBOL]";
-        my $badpart  = $color ?? "\e[31m" !! "";
-        my $clearcol = $color ?? "\e[0m" !! "";
+        my ($red,$clear,$green,$,$eject) = Rakudo::Internals.error-rcgye;
         my sub escape($str) { $str.perl.substr(1).chop }
         join '', "in '",
-                $goodpart,
+                $green,
                 escape(substr($.source,0, $.pos)),
-                $ejection,
-                $badpart,
+                $eject,
+                $red,
                 escape(substr($.source,$.pos)),
-                $clearcol,
+                $clear,
                 "' (indicated by ",
-                $ejection,
-                $clearcol,
+                $eject,
+                $clear,
                 ")",
                 ;
     }
@@ -1675,6 +1756,36 @@ my class X::Str::Trans::InvalidArg is Exception {
     }
 }
 
+my class X::Str::Sprintf::Directives::Count is Exception {
+    has $.args-used;
+    has $.args-have;
+    method message() {
+        "Your printf-style directives specify "
+        ~ ($.args-used == 1 ?? "1 argument, but "
+                            !! "$.args-used arguments, but ")
+        ~ ($.args-have < 1      ?? "no argument was"
+            !! $.args-have == 1 ?? "1 argument was"
+                                !! "$.args-have arguments were")
+        ~ " supplied";
+    }
+}
+
+my class X::Str::Sprintf::Directives::Unsupported is Exception {
+    has $.directive;
+    has $.sequence;
+    method message() {
+        "Directive $.directive is not valid in sprintf format sequence $.sequence"
+    }
+}
+
+my class X::Str::Sprintf::Directives::BadType is Exception {
+    has $.type;
+    has $.directive;
+    method message() {
+        "Directive $.directive not applicable for type $.type"
+    }
+}
+
 my class X::Range::InvalidArg is Exception {
     has $.got is default(Nil);
     method message() {
@@ -1751,8 +1862,20 @@ my class X::TypeCheck is Exception {
     has $.operation;
     has $.got is default(Nil);
     has $.expected is default(Nil);
-    method gotn()      { (try $!got.^name eq $!expected.^name ?? $!got.perl      !! $!got.^name)      // "?" }
-    method expectedn() { (try $!got.^name eq $!expected.^name ?? $!expected.perl !! $!expected.^name) // "?" }
+    method gotn() {
+        my $perl = (try $!got.perl) // "?";
+        $perl = "$perl.substr(0,21)..." if $perl.chars > 24;
+        (try $!got.^name eq $!expected.^name
+          ?? $perl
+          !! "$!got.^name() ($perl)"
+        ) // "?"
+    }
+    method expectedn() {
+        (try $!got.^name eq $!expected.^name
+          ?? $!expected.perl
+          !! $!expected.^name
+        ) // "?"
+    }
     method priors() {
         my $prior = do if nqp::isconcrete($!got) && $!got ~~ Failure {
             "Earlier failure:\n " ~ $!got.mess ~ "\nFinal error:\n ";
@@ -2010,6 +2133,14 @@ my class X::Numeric::DivideByZero is Exception {
     }
 }
 
+my class X::Numeric::Overflow is Exception {
+    method message() { "Numeric overflow" }
+}
+
+my class X::Numeric::Underflow is Exception {
+    method message() { "Numeric underflow" }
+}
+
 my class X::Numeric::Confused is Exception {
     has $.num;
     has $.base;
@@ -2090,7 +2221,7 @@ my class X::Multi::NoMatch is Exception {
         @priors = flat "Earlier failures:\n", @priors, "\nFinal error:\n " if @priors;
         @priors.join ~
         join "\n    ",
-            "Cannot call $.dispatcher.name()$cap; none of these signatures match:",
+            "Cannot resolve caller $.dispatcher.name()$cap; none of these signatures match:",
             @cand;
     }
 }
@@ -2127,52 +2258,65 @@ my class X::PhaserExceptions is Exception {
     }
 }
 
-{
-    my %c_ex;
-    %c_ex{'X::TypeCheck::Binding'} := sub (Mu $got, Mu $expected, $symbol?) {
-            X::TypeCheck::Binding.new(:$got, :$expected, :$symbol).throw;
-        };
-    %c_ex<X::TypeCheck::Assignment> := sub (Mu $symbol, Mu $got, Mu $expected) {
-            X::TypeCheck::Assignment.new(:$symbol, :$got, :$expected).throw;
-        };
-    %c_ex{'X::TypeCheck::Return'} := sub (Mu $got, Mu $expected) {
-            X::TypeCheck::Return.new(:$got, :$expected).throw;
-        };
-    %c_ex<X::Assignment::RO> := sub ($typename = "value") {
-            X::Assignment::RO.new(:$typename).throw;
-        };
-    %c_ex{'X::ControlFlow::Return'} := sub () {
-            X::ControlFlow::Return.new().throw;
-        };
-    %c_ex{'X::NoDispatcher'} := sub ($redispatcher) {
-            X::NoDispatcher.new(:$redispatcher).throw;
-        };
-    %c_ex{'X::Multi::Ambiguous'} := sub ($dispatcher, @ambiguous, $capture) {
-            X::Multi::Ambiguous.new(:$dispatcher, :@ambiguous, :$capture).throw
-        };
-    %c_ex{'X::Multi::NoMatch'} := sub ($dispatcher, $capture) {
-            X::Multi::NoMatch.new(:$dispatcher, :$capture).throw
-        };
-    %c_ex{'X::Role::Initialization'} := sub ($role) {
-            X::Role::Initialization.new(:$role).throw
-        }
-    %c_ex{'X::Role::Parametric::NoSuchCandidate'} := sub (Mu $role) {
-        X::Role::Parametric::NoSuchCandidate.new(:$role).throw;
-        }
-    %c_ex{'X::Inheritance::NotComposed'} = sub ($child-name, $parent-name) {
-        X::Inheritance::NotComposed.new(:$child-name, :$parent-name).throw;
-    }
-    %c_ex{'X::Parameter::RW'} := sub (Mu $got, $symbol) {
-            X::Parameter::RW.new(:$got, :$symbol).throw;
-        };
-    %c_ex{'X::PhaserExceptions'} := sub (@exceptions) {
-            X::PhaserExceptions.new(exceptions =>
-                @exceptions.map(-> Mu \e { EXCEPTION(e) })).throw;
-        };
-    nqp::bindcurhllsym('P6EX', nqp::getattr(%c_ex, Map, '$!storage'));
-
-    0;
-}
+nqp::bindcurhllsym('P6EX', nqp::hash(
+  'X::TypeCheck::Binding', 
+  sub (Mu $got, Mu $expected, $symbol?) {
+      X::TypeCheck::Binding.new(:$got, :$expected, :$symbol).throw;
+  },
+  'X::TypeCheck::Assignment',
+  sub (Mu $symbol, Mu $got, Mu $expected) {
+      X::TypeCheck::Assignment.new(:$symbol, :$got, :$expected).throw;
+  },
+  'X::TypeCheck::Return',
+  sub (Mu $got, Mu $expected) {
+      X::TypeCheck::Return.new(:$got, :$expected).throw;
+  },
+  'X::Assignment::RO',
+  sub ($typename = "value") {
+      X::Assignment::RO.new(:$typename).throw;
+  },
+  'X::ControlFlow::Return',
+  sub () {
+      X::ControlFlow::Return.new().throw;
+  },
+  'X::NoDispatcher',
+  sub ($redispatcher) {
+      X::NoDispatcher.new(:$redispatcher).throw;
+  },
+  'X::Method::NotFound',
+  sub ($invocant, $method, $typename, $private = False) {
+      X::Method::NotFound.new(:$invocant, :$method, :$typename, :$private).throw
+  },
+  'X::Multi::Ambiguous',
+  sub ($dispatcher, @ambiguous, $capture) {
+      X::Multi::Ambiguous.new(:$dispatcher, :@ambiguous, :$capture).throw
+  },
+  'X::Multi::NoMatch',
+  sub ($dispatcher, $capture) {
+      X::Multi::NoMatch.new(:$dispatcher, :$capture).throw
+  },
+  'X::Role::Initialization',
+  sub ($role) {
+      X::Role::Initialization.new(:$role).throw
+  },
+  'X::Role::Parametric::NoSuchCandidate',
+  sub (Mu $role) {
+      X::Role::Parametric::NoSuchCandidate.new(:$role).throw;
+  },
+  'X::Inheritance::NotComposed',
+  sub ($child-name, $parent-name) {
+      X::Inheritance::NotComposed.new(:$child-name, :$parent-name).throw;
+  },
+  'X::Parameter::RW',
+  sub (Mu $got, $symbol) {
+      X::Parameter::RW.new(:$got, :$symbol).throw;
+  },
+  'X::PhaserExceptions',
+  sub (@exceptions) {
+      X::PhaserExceptions.new(exceptions =>
+        @exceptions.map(-> Mu \e { EXCEPTION(e) })).throw;
+  },
+));
 
 my class X::HyperWhatever::Multiple is Exception {
     method message() {
@@ -2230,6 +2374,12 @@ my class X::TooLateForREPR is X::Comp  {
     }
 }
 
+my class X::MustBeParametric is Exception {
+    has $.type;
+    method message() {
+        "$!type.^name() *must* be parameterized";
+    }
+}
 my class X::NotParametric is Exception {
     has $.type;
     method message() {
@@ -2258,7 +2408,9 @@ my class X::InvalidTypeSmiley does X::Comp {
 
 my class X::Seq::Consumed is Exception {
     method message() {
-        "This Seq has already been iterated, and its values consumed"
+        "This Seq has already been iterated, and its values consumed\n" ~
+        "(you might solve this by adding .cache on usages of the Seq, or\n" ~
+        "by assigning the Seq into an array)"
     }
 }
 
@@ -2290,6 +2442,22 @@ my class X::NotEnoughDimensions is Exception {
     }
 }
 
+my class X::TooManyDimensions is Exception {
+    has $.operation;
+    has $.got-dimensions;
+    has $.needed-dimensions;
+    method message() {
+        "Cannot $.operation a $.needed-dimensions dimension array with $.got-dimensions dimensions"
+    }
+}
+
+my class X::IllegalDimensionInShape is Exception {
+    has $.dim;
+    method message() {
+        "Illegal dimension in shape: $.dim. All dimensions must be integers bigger than 0"
+    }
+}
+
 my class X::Assignment::ArrayShapeMismatch is Exception {
     has $.target-shape;
     has $.source-shape;
@@ -2319,4 +2487,27 @@ my class X::Proc::Unsuccessful is Exception {
     }
 }
 
+class CompUnit::DependencySpecification { ... }
+my class X::CompUnit::UnsatisfiedDependency is Exception {
+    has CompUnit::DependencySpecification $.specification;
+
+    my sub is-core($name) {
+        my @parts = $name.split("::");
+        my $ns := ::CORE.WHO;
+        for @parts {
+            return False unless $ns{$_}:exists;
+            $ns := $ns{$_}.WHO;
+        };
+        True
+    };
+
+    method message() {
+        my $name = $.specification.short-name;
+        my $line = $.specification.source-line-number;
+        is-core($name)
+            ?? "{$name} is a builtin type. You can use it without loading a module."
+            !! "Could not find $.specification at line $line in:\n" ~ $*REPO.repo-chain.map(*.Str).join("\n").indent(4)
+    }
+}
+
 # vim: ft=perl6 expandtab sw=4
