@@ -5,31 +5,30 @@ my class Failure is Nil {
     has int $!handled;
 #?endif
 #?if jvm
-    has $!handled;
+    has Int $!handled;   # alas, native int breaks on the JVM
 #?endif
 
+    method !SET-SELF($!exception) {
+        $!backtrace = $!exception.backtrace || Backtrace.new(5);
+        $!exception.reset-backtrace;
+        self
+    }
+
     multi method new() {
         my $stash := CALLER::;
         my $payload = $stash<$!>.DEFINITE ?? $stash<$!> !! "Failed";
-        self.bless(:exception( $payload ~~ Exception
-          ?? $payload !! X::AdHoc.new(:$payload)
-        ))
+        nqp::create(self)!SET-SELF(
+          $payload ~~ Exception ?? $payload !! X::AdHoc.new(:$payload)
+        )
     }
-    multi method new(Exception $exception) {
-        self.bless(:$exception);
+    multi method new(Exception:D \exception) {
+        nqp::create(self)!SET-SELF(exception)
     }
     multi method new($payload) {
-        self.bless(:exception( $payload ~~ Exception
-          ?? $payload !! X::AdHoc.new(:$payload)
-        ))
+        nqp::create(self)!SET-SELF(X::AdHoc.new(:$payload))
     }
     multi method new(|cap (*@msg)) {
-         self.bless(:exception(X::AdHoc.from-slurpy(|cap)));
-    }
-
-    submethod BUILD (:$!exception) {
-        $!backtrace = $!exception.backtrace() || Backtrace.new(8);
-        $!exception.reset-backtrace;
+        nqp::create(self)!SET-SELF(X::AdHoc.from-slurpy(|cap))
     }
 
     # "Shouldn't happen."  We use note here because the dynamic scope in GC is likely meaningless.
@@ -54,7 +53,19 @@ my class Failure is Nil {
         Bool::False;
     }
     multi method Bool(Failure:D:) { $!handled = 1; Bool::False; }
-
+    method handled() {
+        Proxy.new(
+          FETCH => {
+#?if moar
+              nqp::p6bool($!handled)
+#?endif
+#?if jvm
+              $!handled.Bool
+#?endif
+          },
+          STORE => -> $, $value { $!handled = $value.Bool.Numeric }
+       )
+    }
 
 #?if moar
     method Int(Failure:D:)        { $!handled ?? Int !! self!throw(); }
@@ -83,64 +94,58 @@ my class Failure is Nil {
     method FALLBACK(Failure:D: *@) {
         self!throw()
     }
+    method STORE(Failure:D: *@) {
+        self!throw()
+    }
 }
 
 proto sub fail(|) {*};
-multi sub fail() {
+multi sub fail(--> Nil) {
     my $stash := CALLER::;
     my $payload = $stash<$!>.DEFINITE ?? $stash<$!> !! "Failed";
 
     my $fail := Failure.new( $payload ~~ Exception
       ?? $payload !! X::AdHoc.new(:$payload));
 
-    my Mu $return := nqp::getlexcaller('RETURN');
-    $return($fail) unless nqp::isnull($return);
-
-    $fail
+    nqp::throwpayloadlexcaller(nqp::const::CONTROL_RETURN, $fail);
+    CATCH { $fail.exception.throw }
 }
-multi sub fail(Exception:U $e) {
+multi sub fail(Exception:U $e --> Nil) {
     my $fail := Failure.new(
         X::AdHoc.new(:payload("Failed with undefined " ~ $e.^name))
     );
-    my Mu $return := nqp::getlexcaller('RETURN');
-    $return($fail) unless nqp::isnull($return);
-    $fail
+    nqp::throwpayloadlexcaller(nqp::const::CONTROL_RETURN, $fail);
+    CATCH { $fail.exception.throw }
 }
-multi sub fail($payload) {
+multi sub fail($payload --> Nil) {
     my $fail := Failure.new( $payload ~~ Exception
       ?? $payload
       !! X::AdHoc.new(:$payload)
     );
-
-    my Mu $return := nqp::getlexcaller('RETURN');
-    $return($fail) unless nqp::isnull($return);
-
-    $fail
+    nqp::throwpayloadlexcaller(nqp::const::CONTROL_RETURN, $fail);
+    CATCH { $fail.exception.throw }
 }
-multi sub fail(|cap (*@msg)) {
+multi sub fail(|cap (*@msg) --> Nil) {
     my $fail := Failure.new(X::AdHoc.from-slurpy(|cap));
-    my Mu $return := nqp::getlexcaller('RETURN');
-    $return($fail) unless nqp::isnull($return);
-    $fail
+    nqp::throwpayloadlexcaller(nqp::const::CONTROL_RETURN, $fail);
+    CATCH { $fail.exception.throw }
 }
-multi sub fail(Failure:U $f) {
+multi sub fail(Failure:U $f --> Nil) {
     my $fail := Failure.new(
         X::AdHoc.new(:payload("Failed with undefined " ~ $f.^name))
     );
-    my Mu $return := nqp::getlexcaller('RETURN');
-    $return($fail) unless nqp::isnull($return);
-    $fail
+    nqp::throwpayloadlexcaller(nqp::const::CONTROL_RETURN, $fail);
+    CATCH { $fail.exception.throw }
 }
-multi sub fail(Failure:D $fail) {
-    my Mu $return := nqp::getlexcaller('RETURN');
-    $return($fail) unless nqp::isnull($return);
-    $fail
+multi sub fail(Failure:D $fail --> Nil) {
+    nqp::throwpayloadlexcaller(nqp::const::CONTROL_RETURN, $fail);
+    CATCH { $fail.exception.throw }
 }
 
-multi sub die(Failure:D $f) {
+multi sub die(Failure:D $f --> Nil) {
     $f.exception.throw
 }
-multi sub die(Failure:U $f) {
+multi sub die(Failure:U $f --> Nil) {
     X::AdHoc.new(:payload("Died with undefined " ~ $f.^name)).throw;
 }
 
