@@ -22,12 +22,7 @@ my class IO::Handle does IO {
       :$enc   = 'utf8',
       :$nl-in is copy = ["\x0A", "\r\n"],
       Str:D :$nl-out is copy = "\n",
-      :$nl
     ) {
-        if $nl.defined {
-            DEPRECATED(what => ':nl parameter to open', ':nl-in and :nl-out');
-            $nl-in = $nl-out = $nl;
-        }
 
         $mode //= do {
             when so ($r && $w) || $rw { $create              = True; 'rw' }
@@ -94,11 +89,13 @@ my class IO::Handle does IO {
             $truncate  ?? 't' !! '',
             $exclusive ?? 'x' !! '';
 
-        # TODO: catch error, and fail()
-        $!PIO := nqp::open(
-          nqp::unbox_s($!path.abspath),
-          nqp::unbox_s($llmode),
-        );
+        {
+            CATCH { .fail }
+            $!PIO := nqp::open(
+              nqp::unbox_s($!path.abspath),
+              nqp::unbox_s($llmode),
+            );
+        }
 
         $!chomp = $chomp;
         $!nl-out = $nl-out;
@@ -108,24 +105,6 @@ my class IO::Handle does IO {
         self;
     }
 
-    method ins() {
-        DEPRECATED('.lines.kv');
-        0
-    }
-
-    method nl is rw {
-        DEPRECATED('nl-in and nl-out');
-        Proxy.new(
-          FETCH => {
-              $!nl-out
-          },
-          STORE => -> $, $nl {
-            $!nl-out = $nl;
-            Rakudo::Internals.SET_LINE_ENDING_ON_HANDLE($!PIO, $!nl-in = $nl);
-          }
-        );
-    }
-
     method nl-in is rw {
         Proxy.new(
           FETCH => {
@@ -175,47 +154,56 @@ my class IO::Handle does IO {
 
         Seq.new(class :: does Iterator {
             has Mu  $!handle;
-            has Mu  $!size;
+            has int $!size;
             has int $!close;
 
-            submethod BUILD(\handle, \size, \close) {
+            method !SET-SELF(\handle, \size, \close) {
                 $!handle := handle;
                 $!size    = size.Int;
                 $!close   = close;
                 self
             }
             method new(\handle, \size, \close) {
-                nqp::create(self).BUILD(handle, size, close);
+                nqp::create(self)!SET-SELF(handle, size, close);
             }
 
             method pull-one() {
-                my str $str = $!handle.readchars($!size);
-                if nqp::chars($str) {
-                    nqp::p6box_s($str)
-                }
-                else {
-                    $!handle.close if $!close;
+                nqp::if(
+                  nqp::chars(my str $str = $!handle.readchars($!size)),
+                  nqp::p6box_s($str),
+                  nqp::stmts(
+                    nqp::if(
+                      $!close,
+                      $!handle.close
+                    ),
                     IterationEnd
-                }
+                  )
+                )
             }
             method push-all($target) {
                 my str $str = $!handle.readchars($!size);
-                while nqp::chars($str) == $size {
-                    $target.push(nqp::p6box_s($str));
-                    $str = $!handle.readchars($!size);
-                }
+                nqp::while(
+                  nqp::iseq_i(nqp::chars($str),$!size), 
+                  nqp::stmts(
+                    $target.push(nqp::p6box_s($str)),
+                    ($str = $!handle.readchars($!size))
+                  )
+                );
                 $target.push(nqp::p6box_s($str)) if nqp::chars($str);
                 $!handle.close if $!close;
                 IterationEnd
             }
             method count-only() {
-                my int $found;
+                my int $found = 0;
                 my str $str = $!handle.readchars($!size);
-                while nqp::chars($str) == $size {
-                    $found = $found + 1;
-                    $str   = $!handle.readchars($!size);
-                }
-                $found = $found + 1 if nqp::chars($str);
+                nqp::while(
+                  nqp::iseq_i(nqp::chars($str),$!size),
+                  nqp::stmts(
+                    ($found = nqp::add_i($found,1)),
+                    ($str = $!handle.readchars($!size)),
+                  )
+                );
+                ++$found if nqp::chars($str);
                 $!handle.close if $!close;
                 nqp::p6box_i($found)
             }
@@ -236,7 +224,7 @@ my class IO::Handle does IO {
             has int $!elems;
             has int $!done;
 
-            submethod BUILD(\handle, \comber, \close) {
+            method !SET-SELF(\handle, \comber, \close) {
                 $!handle := handle;
                 nqp::istype(comber,Regex)
                   ?? ($!regex := comber)
@@ -247,7 +235,7 @@ my class IO::Handle does IO {
                 self
             }
             method new(\handle, \comber, \close) {
-                nqp::create(self).BUILD(handle, comber, close);
+                nqp::create(self)!SET-SELF(handle, comber, close);
             }
             method !next-chunk(--> Nil) {
                 my int $chars = nqp::chars($!left);
@@ -256,7 +244,7 @@ my class IO::Handle does IO {
                     $!done = 1;
                 }
                 else {
-                    $!strings := nqp::list();
+                    $!strings := nqp::list_s;
                     with $!regex {
                         my \matches   = $!str.match($!regex, :g);
                         $!elems = matches.elems;
@@ -269,7 +257,7 @@ my class IO::Handle does IO {
                             $match := matches[$i];
                             $from = $match.from;
                             $to   = $match.to;
-                            nqp::bindpos($!strings,$i,
+                            nqp::bindpos_s($!strings,$i,
                               nqp::substr($!str,$from,$to - $from));
                             $i = $i + 1;
                         }
@@ -280,7 +268,7 @@ my class IO::Handle does IO {
                         my int $found;
                         my int $extra = nqp::chars($!comber);
                         while ($found = nqp::index($!str,$!comber,$pos)) >= 0 {
-                            nqp::push($!strings,$!comber);
+                            nqp::push_s($!strings,$!comber);
                             $pos = $found + $extra;
                         }
                         $!left  = nqp::substr($!str,$pos);
@@ -292,13 +280,13 @@ my class IO::Handle does IO {
             method pull-one() {
                 if $!elems {
                     $!elems = $!elems - 1;
-                    nqp::p6box_s(nqp::shift($!strings));
+                    nqp::p6box_s(nqp::shift_s($!strings));
                 }
                 else {
                     self!next-chunk until $!elems || $!done;
                     if $!elems {
                         $!elems = $!elems - 1;
-                        nqp::p6box_s(nqp::shift($!strings));
+                        nqp::p6box_s(nqp::shift_s($!strings));
                     }
                     else {
                         $!handle.close if $!close;
@@ -309,7 +297,7 @@ my class IO::Handle does IO {
             method push-all($target) {
                 while $!elems {
                     while $!elems {
-                        $target.push(nqp::p6box_s(nqp::shift($!strings)));
+                        $target.push(nqp::p6box_s(nqp::shift_s($!strings)));
                         $!elems = $!elems - 1;
                     }
                     self!next-chunk until $!elems || $!done;
@@ -341,7 +329,7 @@ my class IO::Handle does IO {
             has int $index;
             has int $chars;
 
-            submethod BUILD(\handle, \close, \COMB) {
+            method !SET-SELF(\handle, \close, \COMB) {
                 $!handle := handle;
                 $!close   = close;
                 $!COMB    = ?COMB;
@@ -350,7 +338,7 @@ my class IO::Handle does IO {
                 self
             }
             method new(\handle, \close, \COMB) {
-                nqp::create(self).BUILD(handle, close, COMB);
+                nqp::create(self)!SET-SELF(handle, close, COMB);
             }
             method !next-chunk(--> Nil) {
                 $!str   = $!handle.readchars;
@@ -414,7 +402,7 @@ my class IO::Handle does IO {
             has int $!elems;
             has int $!done;
 
-            submethod BUILD(\handle, \splitter, \close) {
+            method !SET-SELF(\handle, \splitter, \close) {
                 $!handle := handle;
                 nqp::istype(splitter,Regex)
                   ?? ($!regex   := splitter)
@@ -425,7 +413,7 @@ my class IO::Handle does IO {
                 self
             }
             method new(\handle, \splitter, \close) {
-                nqp::create(self).BUILD(handle, splitter, close);
+                nqp::create(self)!SET-SELF(handle, splitter, close);
             }
             method !next-chunk(--> Nil) {
                 my int $chars = nqp::chars($!left);
@@ -517,7 +505,7 @@ my class IO::Handle does IO {
             has int $!pos;
             has int $!searching;
 
-            submethod BUILD(\handle, $!close) {
+            method !SET-SELF(\handle, $!close) {
                 $!handle   := handle;
                 $!searching = 1;
                 $!str       = ""; # RT #126492
@@ -525,7 +513,7 @@ my class IO::Handle does IO {
                 self
             }
             method new(\handle, \close) {
-                nqp::create(self).BUILD(handle, close);
+                nqp::create(self)!SET-SELF(handle, close);
             }
             method !next-chunk() {
                 my int $chars = nqp::chars($!str);
@@ -636,12 +624,12 @@ my class IO::Handle does IO {
             has $!handle;
             has $!close;
 
-            submethod BUILD(\handle, $!close) {
+            method !SET-SELF(\handle, $!close) {
                 $!handle := handle;
                 self
             }
             method new(\handle, \close) {
-                nqp::create(self).BUILD(handle, close);
+                nqp::create(self)!SET-SELF(handle, close);
             }
             method pull-one() is raw {
                 $!handle.get // do {
@@ -671,7 +659,7 @@ my class IO::Handle does IO {
         $buf;
     }
 
-    method readchars(Int(Cool:D) $chars = 65536) { # optimize for ASCII
+    method readchars(Int(Cool:D) $chars = $*DEFAULT-READ-ELEMS) {
 #?if jvm
         my Buf $buf := Buf.new;   # nqp::readcharsfh doesn't work on the JVM
         # a char = 2 bytes
@@ -683,14 +671,17 @@ my class IO::Handle does IO {
 #?endif
     }
 
-    method Supply(IO::Handle:D: :$size = 65536, :$bin --> Supply:D) {
+    method Supply(IO::Handle:D: :$size = $*DEFAULT-READ-ELEMS, :$bin --> Supply:D) {
         if $bin {
             supply {
                 my $buf := self.read($size);
-                while nqp::elems($buf) {
-                    emit $buf;
-                    $buf := self.read($size);
-                }
+                nqp::while(
+                  nqp::elems($buf),
+                  nqp::stmts(
+                    (emit $buf),
+                    ($buf := self.read($size))
+                  )
+                );
                 done;
             }
         }
@@ -698,26 +689,21 @@ my class IO::Handle does IO {
             supply {
                 my int $chars = $size;
                 my str $str = self.readchars($chars);
-                while nqp::chars($str) {
-                    emit nqp::p6box_s($str);
-                    $str = self.readchars($chars);
-                }
+                nqp::while(
+                  nqp::chars($str),
+                  nqp::stmts(
+                    (emit nqp::p6box_s($str)),
+                    ($str = self.readchars($chars))
+                  )
+                );
                 done;
             }
         }
     }
 
     proto method seek(|) { * }
-    multi method seek(IO::Handle:D: Int:D $offset, Int:D $whence --> True) {
-        DEPRECATED(
-          SeekType.^enum_value_list[$whence],
-          :what("numerical seektype $whence"),
-        );
-        nqp::seekfh($!PIO, $offset, $whence);
-    }
-    multi method seek(IO::Handle:D: Int:D $offset, SeekType:D $whence) {
+    multi method seek(IO::Handle:D: Int:D $offset, SeekType:D $whence = SeekFromBeginning) {
         nqp::seekfh($!PIO, $offset, +$whence);
-        True;
     }
 
     method tell(IO::Handle:D:) returns Int {
@@ -733,7 +719,15 @@ my class IO::Handle does IO {
     }
 
     method t(IO::Handle:D:) {
-        self.opened && nqp::p6bool($!PIO.isatty)
+        self.opened && nqp::p6bool(nqp::isttyfh($!PIO))
+    }
+
+    method lock(IO::Handle:D: Int:D $flag) {
+        nqp::lockfh($!PIO, $flag)
+    }
+
+    method unlock(IO::Handle:D: --> True) {
+        nqp::unlockfh($!PIO);
     }
 
 
@@ -774,17 +768,16 @@ my class IO::Handle does IO {
     }
 
     proto method slurp-rest(|) { * }
-    multi method slurp-rest(IO::Handle:D: :$bin!) returns Buf {
-        my $Buf := buf8.new();
+    multi method slurp-rest(IO::Handle:D: :$bin! where *.so) returns Buf {
+        my $res := buf8.new();
         loop {
-            my $buf := buf8.new();
-            nqp::readfh($!PIO,$buf,65536);
-            last if $buf.bytes == 0;
-            $Buf := $Buf ~ $buf;
+            my $buf := nqp::readfh($!PIO,buf8.new,0x100000);
+            last unless nqp::elems($buf);
+            $res.push($buf);
         }
-        $Buf;
+        $res
     }
-    multi method slurp-rest(IO::Handle:D: :$enc) returns Str {
+    multi method slurp-rest(IO::Handle:D: :$enc, :$bin) returns Str {
         self.encoding($enc) if $enc.defined;
         nqp::p6box_s(nqp::readallfh($!PIO));
     }
@@ -796,12 +789,12 @@ my class IO::Handle does IO {
 
     multi method gist(IO::Handle:D:) {
         self.opened
-            ?? "IO::Handle<$!path>(opened, at octet {$.tell})"
-            !! "IO::Handle<$!path>(closed)"
+            ?? self.^name ~ "<$!path.gist()>(opened, at octet {$.tell})"
+            !! self.^name ~ "<$!path.gist()>(closed)"
     }
 
     multi method perl(IO::Handle:D:) {
-        "IO::Handle.new(path => {$!path.perl}, chomp => {$!chomp.perl})"
+        self.^name ~ ".new({:$!path.perl},{$!chomp ?? :$!chomp.perl !! ''})"
     }
 
 
@@ -833,12 +826,21 @@ my class IO::Handle does IO {
     method modified(IO::Handle:D:) { $!path.modified }
     method accessed(IO::Handle:D:) { $!path.accessed }
     method changed(IO::Handle:D:)  { $!path.changed  }
+    method mode(IO::Handle:D:)     { $!path.mode     }
 
 #?if moar
     method watch(IO::Handle:D:) {
         IO::Notification.watch-path($!path);
     }
 #?endif
+
+    method native-descriptor(IO::Handle:D:) {
+        nqp::filenofh($!PIO)
+    }
+}
+
+Rakudo::Internals.REGISTER-DYNAMIC: '$*DEFAULT-READ-ELEMS', {
+    PROCESS::<$DEFAULT-READ-ELEMS> := %*ENV<RAKUDO_DEFAULT_READ_ELEMS> // 65536;
 }
 
 # vim: ft=perl6 expandtab sw=4
