@@ -17,7 +17,7 @@ my class IO::Path is Cool {
         nqp::p6bool(nqp::iseq_s($.abspath,nqp::unbox_s(IO::Path.new(|that).abspath)));
     }
 
-    submethod BUILD(Str() :$!path!, :$!SPEC!, Str() :$!CWD!) { }
+    submethod BUILD(Str() :$!path!, :$!SPEC!, Str() :$!CWD! --> Nil) { }
 
     method new-from-absolute-path($path, :$SPEC = $*SPEC, :$CWD = $*CWD) {
         method !fap() {
@@ -47,7 +47,7 @@ my class IO::Path is Cool {
     }
 
     method abspath() {
-        $!abspath //= substr($!path,0,1) eq '-'
+        $!abspath //= $!path.starts-with('-')
           ?? ''
           !! $!SPEC.rel2abs($!path,$!CWD);
     }
@@ -64,7 +64,7 @@ my class IO::Path is Cool {
     method volume(IO::Path:D:)   { %.parts<volume>   }
     method dirname(IO::Path:D:)  { %.parts<dirname>  }
     method basename(IO::Path:D:) { %.parts<basename> }
-    method extension(IO::Path:D:) { MAKE-EXT(self.basename) }
+    method extension(IO::Path:D:) { Rakudo::Internals.MAKE-EXT(self.basename) }
 
     # core can't do 'basename handles <Numeric Bridge Int>'
     method Numeric(IO::Path:D:) { self.basename.Numeric }
@@ -73,13 +73,14 @@ my class IO::Path is Cool {
 
     multi method Str (IO::Path:D:) { $!path }
     multi method gist(IO::Path:D:) {
-        qq|"$.abspath".IO|;
+        $!is-absolute
+          ?? qq|"$.abspath".IO|
+          !! qq|"$.path".IO|
     }
     multi method perl(IO::Path:D:) {
-        ($.is-absolute
-          ?? "q|$.abspath|.IO(:SPEC({$!SPEC.^name}))"
-          !! "q|$.path|.IO(:SPEC({$!SPEC.^name}),:CWD<$!CWD>)"
-        ).subst(:global, '\\', '\\\\');
+        $!is-absolute  # attribute now set
+          ?? "{$.abspath.perl}.IO({:$!SPEC.perl})"
+          !! "{$.path.perl}.IO({:$!SPEC.perl},{:$!CWD.perl})"
     }
 
     method succ(IO::Path:D:) {
@@ -382,7 +383,7 @@ my class IO::Path is Cool {
         CATCH { default {
             fail X::IO::Mkdir.new(:path($!abspath), :$mode, os-error => .Str);
         } }
-        True;
+        $!e = True;
     }
 
     method rmdir(IO::Path:D:) {
@@ -390,6 +391,7 @@ my class IO::Path is Cool {
         CATCH { default {
             fail X::IO::Rmdir.new(:path($!abspath), os-error => .Str);
         } }
+        $!e = False;
         True;
     }
 
@@ -402,30 +404,31 @@ my class IO::Path is Cool {
 
         CATCH { default {
             fail X::IO::Dir.new(
-              :path(nqp::box_s($.abspath,Str)), :os-error(.Str) );
+              :path($.abspath), :os-error(.Str) );
         } }
-        my $cwd_chars = $CWD.chars;
 
-#?if moar
-        my str $cwd = nqp::cwd();
-        nqp::chdir(nqp::unbox_s($.abspath));
-#?endif
-        my $abspath-sep := $.abspath eq $!SPEC.dir-sep
-          ?? $!SPEC.dir-sep
+        my str $abspath = $.abspath.ends-with($!SPEC.dir-sep)
+          ?? $.abspath
           !! $.abspath ~ $!SPEC.dir-sep;
 
+        my str $path = $.path eq '.' | $!SPEC.dir-sep
+          ?? ''
+          !! $.path.ends-with($!SPEC.dir-sep)
+            ?? $.path
+            !! $.path ~ $!SPEC.dir-sep;
+
         my Mu $dirh := nqp::opendir(nqp::unbox_s($.abspath));
         gather {
 #?if jvm
             for <. ..> -> $elem {
                 if $test.ACCEPTS($elem) {
                     $Str
-                      ?? $absolute
-                        ?? take $abspath-sep ~ $elem
-                        !! take substr($abspath-sep ~ $elem,$cwd_chars + 1)
-                      !! $absolute
-                        ?? take IO::Path.new-from-absolute-path($abspath-sep ~ $elem,:$!SPEC,:$CWD)
-                        !! take substr($abspath-sep ~ $elem,$cwd_chars + 1).IO(:$!SPEC,:$CWD);
+                      ?? !$absolute
+                        ?? take $path ~ $elem
+                        !! take $abspath ~ $elem
+                      !! !$absolute
+                        ?? take IO::Path.new($path ~ $elem,:$!SPEC,:$CWD)
+                        !! take IO::Path.new-from-absolute-path($abspath ~ $elem,:$!SPEC,:$CWD);
                 }
             }
 #?endif
@@ -435,25 +438,16 @@ my class IO::Path is Cool {
                     nqp::closedir($dirh);
                     last;
                 }
-                my Str $elem = nqp::box_s($str_elem,Str);
-#?if jvm
-                if $test.ACCEPTS($!SPEC.basename($elem)) {
-#?endif
-#?if !jvm
-                if $test.ACCEPTS($elem) {
-                    $elem = $abspath-sep ~ $elem; # make absolute
-#?endif
+
+                if $test.ACCEPTS($str_elem) {
                     $Str
                       ?? !$absolute && !$.is-absolute
-                        ?? take substr($elem,$cwd_chars + 1)
-                        !! take $elem
+                        ?? take $path ~ $str_elem
+                        !! take $abspath ~ $str_elem
                       !! !$absolute && !$.is-absolute
-                        ?? take substr($elem,$cwd_chars + 1).IO(:$!SPEC,:$CWD)
-                        !! take IO::Path.new-from-absolute-path($elem,:$!SPEC,:$CWD);
+                        ?? take IO::Path.new($path ~ $str_elem,:$!SPEC,:$CWD)
+                        !! take IO::Path.new-from-absolute-path($abspath ~ $str_elem,:$!SPEC,:$CWD);
                 }
-#?if moar
-                nqp::chdir($cwd);
-#?endif
             }
         }
     }
@@ -465,21 +459,27 @@ my class IO::Path is Cool {
 
         my Mu $PIO := nqp::getattr(nqp::decont($handle),IO::Handle,'$!PIO');
         if $bin {
-            my $Buf := buf8.new();
-            loop {
-                my $buf := buf8.new();
-                nqp::readfh($PIO,$buf,65536);
-                last if $buf.bytes == 0;
-                $Buf := $Buf ~ $buf;
+            # normal file
+            if Rakudo::Internals.FILETEST-S(self.abspath) -> int $size {
+                nqp::readfh($PIO,buf8.new,$size)
+            }
+            # spooky file with zero size?
+            else {
+                my $res := buf8.new();
+                loop {
+                    my $buf := nqp::readfh($PIO,buf8.new,0x100000);
+                    last unless nqp::elems($buf);
+                    $res.push($buf);
+                }
+                $handle.close;
+                $res
             }
-            $handle.close;
-            $Buf;
         }
         else {
             $handle.encoding($enc) if $enc.defined;
             my $slurped := nqp::p6box_s(nqp::readallfh($PIO));
             $handle.close;
-            $slurped;
+            $slurped
         }
     }
 
@@ -506,139 +506,124 @@ my class IO::Path is Cool {
 
     proto method lines(|) { * }
     multi method lines(IO::Path:D: |c) {
-        my $handle = self.open(|c);
-        $handle && $handle.lines(:close);
+        self.open(|c).lines(:close);
     }
 
     proto method comb(|) { * }
     multi method comb(IO::Path:D: Cool:D $comber = "", |c) {
-        my $handle = self.open(|c);
-        $handle && $handle.comb($comber, :close);
+        self.open(|c).comb($comber, :close);
     }
     multi method comb(IO::Path:D: Int:D $size, |c) {
-        my $handle = self.open(|c);
-        $handle && $handle.comb($size, :close);
+        self.open(|c).comb($size, :close);
     }
     multi method comb(IO::Path:D: Regex:D $comber, |c) {
-        my $handle = self.open(|c);
-        $handle && $handle.comb($comber, :close);
+        self.open(|c).comb($comber, :close);
     }
 
     multi method split(IO::Path:D: Str:D $splitter = "", |c) {
-        my $handle = self.open(|c);
-        $handle && $handle.split($splitter, :close);
+        self.open(|c).split($splitter, :close);
     }
     multi method split(IO::Path:D: Regex:D $splitter, |c) {
-        my $handle = self.open(|c);
-        $handle && $handle.split($splitter, :close);
+        self.open(|c).split($splitter, :close);
     }
 
     proto method words(|) { * }
     multi method words(IO::Path:D: |c) {
-        my $handle = self.open(|c);
-        $handle && $handle.words(:close);
+        self.open(|c).words(:close);
     }
 
-    my %t =
-      e => -> $p { True }, # if we get here, it exists
-      d => -> $p { nqp::p6bool(nqp::stat(nqp::unbox_s($p),nqp::const::STAT_ISDIR)) },
-      f => -> $p { nqp::p6bool(nqp::stat(nqp::unbox_s($p),nqp::const::STAT_ISREG)) },
-      s => -> $p { %t.AT-KEY("f")($p) && nqp::box_i(nqp::stat(nqp::unbox_s($p),nqp::const::STAT_FILESIZE),Int) },
-      l => -> $p { nqp::p6bool(nqp::fileislink(nqp::unbox_s($p))) },
-      r => -> $p { nqp::p6bool(nqp::filereadable(nqp::unbox_s($p))) },
-      w => -> $p { nqp::p6bool(nqp::filewritable(nqp::unbox_s($p))) },
-      x => -> $p { nqp::p6bool(nqp::fileexecutable(nqp::unbox_s($p))) },
-      z => -> $p { %t.AT-KEY("f")($p) && nqp::p6bool(nqp::stat(nqp::unbox_s($p),nqp::const::STAT_FILESIZE) == 0) },
-
-      "!e" => -> $p { False }, # if we get here, it exists
-      "!d" => -> $p { !%t.AT-KEY("d")($p) },
-      "!f" => -> $p { !%t.AT-KEY("f")($p) },
-      "!l" => -> $p { !%t.AT-KEY("l")($p) },
-      "!r" => -> $p { !%t.AT-KEY("r")($p) },
-      "!w" => -> $p { !%t.AT-KEY("w")($p) },
-      "!x" => -> $p { !%t.AT-KEY("x")($p) },
-      "!z" => -> $p { !%t.AT-KEY("z")($p) },
-    ;
-
-    method all(*@tests) {
-        return False if !@tests or !$.e;
-
-        my $result = True;
-        for @tests -> $t {
-            die "Unknown test $t" unless %t.EXISTS-KEY($t);
-            last unless $result = $result && %t.AT-KEY($t)($!abspath);
-        }
-
-        $result;
+    method e(--> Bool) {
+        $!e //= ?Rakudo::Internals.FILETEST-E($.abspath) # must be $.abspath
+    }
+    method d(--> Bool) {
+        $.e
+          ?? ?Rakudo::Internals.FILETEST-D($!abspath)
+          !! fail X::IO::DoesNotExist.new(:path(~self),:trying<d>)
     }
 
-    method e() { $!e //= FILETEST-E($.abspath) }
-
-    method d() {
-        fail X::IO::DoesNotExist.new(:path(self.Str),:trying<d>) if !$.e;
-        FILETEST-D($!abspath);
+    method f(--> Bool) {
+        $.e
+          ?? ?Rakudo::Internals.FILETEST-F($!abspath)
+          !! fail X::IO::DoesNotExist.new(:path(~self),:trying<f>)
     }
 
-    method f() {
-        fail X::IO::DoesNotExist.new(:path(self.Str),:trying<f>) if !$.e;
-        FILETEST-F($!abspath);
+    method s(--> Int) {
+        $.e
+          ?? $.f
+            ?? Rakudo::Internals.FILETEST-S($!abspath)
+            !! fail X::IO::NotAFile.new(:path(~self),:trying<s>)
+          !! fail X::IO::DoesNotExist.new(:path(~self),:trying<s>)
     }
 
-    method s() {
-        fail X::IO::DoesNotExist.new(:path(self.Str),:trying<s>) if !$.e;
-        fail X::IO::NotAFile.new(:path(self.Str),:trying<s>)     if !$.f;
-        FILETEST-S($!abspath);
+    method l(--> Bool) {
+        $.e
+          ?? ?Rakudo::Internals.FILETEST-L($!abspath)
+          !! fail X::IO::DoesNotExist.new(:path(~self),:trying<l>)
     }
 
-    method l() {
-        fail X::IO::DoesNotExist.new(:path(self.Str),:trying<l>) if !$.e;
-        FILETEST-L($!abspath);
+    method r(--> Bool) {
+        $.e
+          ?? ?Rakudo::Internals.FILETEST-R($!abspath)
+          !! fail X::IO::DoesNotExist.new(:path(~self),:trying<r>)
     }
 
-    method r() {
-        fail X::IO::DoesNotExist.new(:path(self.Str),:trying<r>) if !$.e;
-        FILETEST-R($!abspath);
+    method w(--> Bool) {
+        $.e
+          ?? ?Rakudo::Internals.FILETEST-W($!abspath)
+          !! fail X::IO::DoesNotExist.new(:path(~self),:trying<w>)
     }
 
-    method w() {
-        fail X::IO::DoesNotExist.new(:path(self.Str),:trying<w>) if !$.e;
-        FILETEST-W($!abspath);
+    method rw(--> Bool) {
+        $.e
+          ?? Rakudo::Internals.FILETEST-RW($!abspath)
+          !! fail X::IO::DoesNotExist.new(:path(~self),:trying<rw>)
     }
 
-    method rw() {
-        fail X::IO::DoesNotExist.new(:path(self.Str),:trying<w>) if !$.e;
-        FILETEST-RW($!abspath);
+    method x(--> Bool) {
+        $.e
+          ?? ?Rakudo::Internals.FILETEST-X($!abspath)
+          !! fail X::IO::DoesNotExist.new(:path(~self),:trying<x>)
     }
 
-    method x() {
-        fail X::IO::DoesNotExist.new(:path(self.Str),:trying<x>) if !$.e;
-        FILETEST-X($!abspath);
+    method rwx(--> Bool) {
+        $.e
+          ?? ?Rakudo::Internals.FILETEST-RWX($!abspath)
+          !! fail X::IO::DoesNotExist.new(:path(~self),:trying<rwx>)
     }
 
-    method rwx() {
-        fail X::IO::DoesNotExist.new(:path(self.Str),:trying<w>) if !$.e;
-        FILETEST-RWX($!abspath);
+    method z(--> Bool) {
+        $.e
+          ?? $.f
+            ?? ?Rakudo::Internals.FILETEST-Z($!abspath)
+            !! fail X::IO::NotAFile.new(:path(~self),:trying<z>)
+          !! fail X::IO::DoesNotExist.new(:path(~self),:trying<z>)
     }
 
-    method z() {
-        fail X::IO::DoesNotExist.new(:path(self.Str),:trying<z>) if !$.e;
-        fail X::IO::NotAFile.new(:path(self.Str),:trying<z>)     if !$.f;
-        FILETEST-Z($!abspath);
+    method modified(--> Instant) {
+        $.e
+          ?? Instant.from-posix(Rakudo::Internals.FILETEST-MODIFIED($!abspath))
+          !! fail X::IO::DoesNotExist.new(:path(~self),:trying<modified>)
     }
 
-    method modified() {
-        fail X::IO::DoesNotExist.new(:path(self.Str),:trying<modified>) if !$.e;
-        FILETEST-MODIFIED($!abspath);
+    method accessed(--> Instant) {
+        $.e
+          ?? Instant.from-posix(Rakudo::Internals.FILETEST-ACCESSED($!abspath))
+          !! fail X::IO::DoesNotExist.new(:path(~self),:trying<accessed>)
     }
 
-    method accessed() {
-        fail X::IO::DoesNotExist.new(:path(self.Str),:trying<accessed>) if !$.e;
-        FILETEST-ACCESSED($!abspath);
+    method changed(--> Instant) {
+        $.e
+          ?? Instant.from-posix(Rakudo::Internals.FILETEST-CHANGED($!abspath))
+          !! fail X::IO::DoesNotExist.new(:path(~self),:trying<changed>)
     }
 
-    method changed() {
-        fail X::IO::DoesNotExist.new(:path(self.Str),:trying<changed>) if !$.e;
-        FILETEST-CHANGED($!abspath);
+    method mode(--> IntStr) {
+        $.e
+          ?? nqp::stmts(
+              (my int $mode = nqp::stat($!abspath, nqp::const::STAT_PLATFORM_MODE) +& 0o7777),
+              IntStr.new($mode, sprintf('%04o', $mode))
+            )
+          !! fail X::IO::DoesNotExist.new(:path(~self),:trying<mode>)
     }
 }
 
