@@ -1,7 +1,11 @@
 class IO::Special {
     has Str $.what;
 
-    submethod BUILD(:$!what!)    { }
+    method new(\what) {
+        my $io := nqp::create(self);
+        nqp::bindattr($io,self,'$!what',what);
+        $io
+    }
     method WHICH(IO::Special:D:) { "IO::Special$!what" }
     method Str(IO::Special:D:)   { $!what }
 
@@ -12,10 +16,11 @@ class IO::Special {
     method f(IO::Special:D:) { False }
     method s(IO::Special:D:) { 0 }
     method l(IO::Special:D:) { False }
-    method r(IO::Special:D:) { $!what eq '<IN>' }
-    method w(IO::Special:D:) { $!what eq '<OUT>' or $!what eq '<ERR>' }
+    method r(IO::Special:D:) { $!what eq '<STDIN>' }
+    method w(IO::Special:D:) { $!what eq '<STDOUT>' or $!what eq '<STDERR>' }
     method x(IO::Special:D:) { False }
     method modified(IO::Special:D:) { Instant }
     method accessed(IO::Special:D:) { Instant }
-    method changed(IO::Special:D:)  { Instant}
+    method changed(IO::Special:D:)  { Instant }
+    method mode(IO::Special:D:)     { Nil     }
 }
