@@ -13,37 +13,62 @@ my sub MAIN_HELPER($retval = 0) {
     my $m = callframe(1).my<&MAIN>;
     return $retval unless $m;
 
+    my $no-named-after = !$*MAIN-ALLOW-NAMED-ANYWHERE; 
+
     # Convert raw command line args into positional and named args for MAIN
     my sub process-cmd-args(@args is copy) {
-        my (@positional-arguments, %named-arguments);
-        my $stopped = False;
-        while +@args {
-            my $passed-value = @args.shift;
-            $stopped = $passed-value eq '--';
-            if !$stopped && $passed-value ~~ /^ ( '--' | '-' | ':' ) ('/'?) (<-[0..9\.]> .*) $/ {
-                my ($switch, $negate, $arg) = (~$0, ?((~$1).chars), ~$2);
+        my $positional := nqp::list;
+        my %named;
 
-                with $arg.index('=') {
-                    my ($name, $value) = $arg.split('=', 2);
-                    $value = val($value);
-                    $value = $value but False if $negate;
-                    %named-arguments.push: $name => $value;
-                } else {
-                    %named-arguments.push: $arg => !$negate;
-                }
-            } else {
-                @args.unshift($passed-value) unless $passed-value eq '--';
-                @positional-arguments.append: @args.map: &val;
+        while ?@args {
+            my str $passed-value = @args.shift;
+
+            # rest considered to be non-parsed
+            if nqp::iseq_s($passed-value,'--') {
+                nqp::push($positional, val($_)) for @args;
                 last;
             }
+
+            # no longer accepting nameds
+            elsif $no-named-after && nqp::isgt_i(nqp::elems($positional),0) {
+                nqp::push($positional, val($passed-value));
+            }
+
+            # named
+            elsif $passed-value
+              ~~ /^ [ '--' | '-' | ':' ] ('/'?) (<-[0..9\.]> .*) $/ {  # 'hlfix
+                my str $arg = $1.Str;
+                my $split  := nqp::split("=",$arg);
+
+                # explicit value
+                if nqp::isgt_i(nqp::elems($split),1) {
+                    my str $name = nqp::shift($split);
+                    %named.push: $name => $0.chars
+                      ?? val(nqp::join("=",$split)) but False
+                      !! val(nqp::join("=",$split));
+                }
+                
+                # implicit value
+                else {
+                    %named.push: $arg => !($0.chars);
+                }
+            }
+
+            # positional
+            else {
+                nqp::push($positional, val($passed-value));
+            }
         }
-        $PROCESS::ARGFILES = IO::ArgFiles.new(:args(@args));
-        @positional-arguments, %named-arguments;
+
+        nqp::p6bindattrinvres(
+          nqp::create(List),List,'$!reified',$positional
+        ),%named;
     }
 
     # Generate $?USAGE string (default usage info for MAIN)
     my sub gen-usage() {
         my @help-msgs;
+        my Pair @arg-help;
 
         my sub strip_path_prefix($name) {
             my $SPEC := $*SPEC;
@@ -65,24 +90,35 @@ my sub MAIN_HELPER($retval = 0) {
           !! strip_path_prefix($*PROGRAM-NAME);
         for $m.candidates -> $sub {
             next if $sub.?is-hidden-from-USAGE;
-            my (@required-named, @optional-named, @positional, $docs);
+
+            my @required-named;
+            my @optional-named;
+            my @positional;
+            my $docs;
+
             for $sub.signature.params -> $param {
                 my $argument;
                 if $param.named {
-                    my @names  = $param.named_names.reverse;
-                    $argument  = @names.map({($^n.chars == 1 ?? '-' !! '--') ~ $^n}).join('|');
-                    $argument ~= "=<{$param.type.^name}>" unless $param.type === Bool;
-                    if $param.optional {
+                    if $param.slurpy {
+                        $argument  = "--<$param.usage-name()>=...";
                         @optional-named.push("[$argument]");
                     }
                     else {
-                        @required-named.push($argument);
+                        my @names  = $param.named_names.reverse;
+                        $argument  = @names.map({($^n.chars == 1 ?? '-' !! '--') ~ $^n}).join('|');
+                        $argument ~= "=<{$param.type.^name}>" unless $param.type === Bool;
+                        if $param.optional {
+                            @optional-named.push("[$argument]");
+                        }
+                        else {
+                            @required-named.push($argument);
+                        }
                     }
                 }
                 else {
                     my $constraints  = $param.constraint_list.map(*.gist).join(' ');
                     my $simple-const = $constraints && $constraints !~~ /^_block/;
-                    $argument = $param.name   ?? '<' ~ substr($param.name,1) ~ '>' !!
+                    $argument = $param.name   ?? "<$param.usage-name()>" !!
                                 $simple-const ??       $constraints                !!
                                                  '<' ~ $param.type.^name     ~ '>' ;
 
@@ -90,6 +126,7 @@ my sub MAIN_HELPER($retval = 0) {
                     $argument = "[$argument]"     if $param.optional;
                     @positional.push($argument);
                 }
+                @arg-help.push($argument => $param.WHY.contents) if $param.WHY and (@arg-help.grep:{ .key eq $argument}) == Empty;  # Use first defined
             }
             if $sub.WHY {
                 $docs = '-- ' ~ $sub.WHY.contents
@@ -97,6 +134,13 @@ my sub MAIN_HELPER($retval = 0) {
             my $msg = join(' ', $prog-name, @required-named, @optional-named, @positional, $docs // '');
             @help-msgs.push($msg);
         }
+
+        if @arg-help {
+            @help-msgs.push('');
+            my $offset = max(@arg-help.map: { .key.chars }) + 4;
+            @help-msgs.append(@arg-help.map: { '  ' ~ .key ~ ' ' x ($offset - .key.chars) ~ .value });
+        }
+
         my $usage = "Usage:\n" ~ @help-msgs.map('  ' ~ *).join("\n");
         $usage;
     }
@@ -153,4 +197,8 @@ my sub MAIN_HELPER($retval = 0) {
     }
 }
 
+Rakudo::Internals.REGISTER-DYNAMIC: '$*MAIN-ALLOW-NAMED-ANYWHERE', {
+    PROCESS::<$MAIN-ALLOW-NAMED-ANYWHERE> := 0;
+}
+
 # vim: ft=perl6 expandtab sw=4
