@@ -29,8 +29,8 @@ my role Numeric {
     method pred() { self - 1 }
 }
 
-multi sub infix:<eqv>(Numeric:D $a, Numeric:D $b) {
-    $a.WHAT === $b.WHAT && ($a cmp $b) == 0
+multi sub infix:<eqv>(Numeric:D \a, Numeric:D \b) {
+    nqp::p6bool(a =:= b || (a.WHAT =:= b.WHAT && a == b)) # RT #127951
 }
 
 ## arithmetic operators
@@ -41,6 +41,8 @@ multi sub prefix:<+>(\a) { a.Numeric }
 proto sub prefix:<->($?) is pure { * }
 multi sub prefix:<->(\a) { -a.Numeric }
 
+sub prefix:<−>($n) is pure { prefix:<->($n) }
+
 proto sub abs($) is pure { * }
 multi sub abs(\a) { abs a.Numeric }
 
@@ -184,34 +186,42 @@ multi sub infix:<+>($x = 0)      { $x.Numeric }
 multi sub infix:<+>(\a, \b)    { a.Numeric + b.Numeric }
 
 proto sub infix:<->(Mu $?, Mu $?) is pure   { * }
-multi sub infix:<->($x = 0)      { $x.Numeric }
+multi sub infix:<->($x = 0)      { -$x.Numeric }
 multi sub infix:<->(\a, \b)    { a.Numeric - b.Numeric }
 
+proto sub infix:<−>(Mu $?, Mu $?) is pure { * }
+multi sub infix:<−>($x = 0)    { -$x.Numeric }
+multi sub infix:<−>(\a, \b) is pure { a.Numeric - b.Numeric }
+
 proto sub infix:<*>(Mu $?, Mu $?) is pure   { * }
 multi sub infix:<*>($x = 1)      { $x.Numeric }
 multi sub infix:<*>(\a, \b)    { a.Numeric * b.Numeric }
 
-proto sub infix:</>(Mu $?, Mu $?) { * }
-multi sub infix:</>()            { fail "No zero-arg meaning for infix:</>" }
+sub infix:<×>(|c) is pure { infix:<*>(|c) }
+
+proto sub infix:</>(Mu $?, Mu $?) is pure { * }
+multi sub infix:</>() { Failure.new("No zero-arg meaning for infix:</>") }
 multi sub infix:</>($x)          { $x.Numeric }
 multi sub infix:</>(\a, \b)    { a.Numeric / b.Numeric }
 
+sub infix:<÷>(|c) is pure { infix:</>(|c) }
+
 proto sub infix:<div>(Mu $?, Mu $?) is pure  { * }
 # rest of infix:<div> is in Int.pm
 
 proto sub infix:<%>(Mu $?, Mu $?) is pure   { * }
-multi sub infix:<%>()            { fail "No zero-arg meaning for infix:<%>" }
+multi sub infix:<%>() { Failure.new("No zero-arg meaning for infix:<%>") }
 multi sub infix:<%>($x)          { $x }
 multi sub infix:<%>(\a, \b)    { a.Real % b.Real }
 
 proto sub infix:<%%>(Mu $?, Mu $?) is pure  { * }
-multi sub infix:<%%>()           { fail "No zero-arg meaning for infix:<%%>" }
+multi sub infix:<%%>() { Failure.new("No zero-arg meaning for infix:<%%>") }
 multi sub infix:<%%>($)         { Bool::True }
 multi sub infix:<%%>(\a, \b)   {
-    fail X::Numeric::DivideByZero.new(
-      using => 'infix:<%%>', numerator => a
-    ) unless b;
-    a.Real % b.Real == 0;
+    b
+      ?? a.Real % b.Real == 0
+      !! Failure.new(X::Numeric::DivideByZero.new(
+           using => 'infix:<%%>', numerator => a))
 }
 
 proto sub infix:<lcm>(Mu $?, Mu $?) is pure  { * }
@@ -219,7 +229,7 @@ multi sub infix:<lcm>(Int $x = 1) { $x }
 multi sub infix:<lcm>(\a, \b)   { a.Int lcm b.Int }
 
 proto sub infix:<gcd>(Mu $?, Mu $?) is pure { * }
-multi sub infix:<gcd>()          { fail 'No zero-arg meaning for infix:<gcd>' }
+multi sub infix:<gcd>() { Failure.new('No zero-arg meaning for infix:<gcd>') }
 multi sub infix:<gcd>(Int $x)    { $x }
 multi sub infix:<gcd>(\a, \b)  { a.Int gcd b.Int }
 
@@ -227,15 +237,32 @@ proto sub infix:<**>(Mu $?, Mu $?) is pure  { * }
 multi sub infix:<**>($x = 1)     { $x.Numeric }
 multi sub infix:<**>(\a, \b)   { a.Numeric ** b.Numeric }
 
+proto sub postfix:<ⁿ>(Mu $, Mu $) is pure  { * }
+multi sub postfix:<ⁿ>(\a, \b)  { a ** b }
+
 ## relational operators
 
-proto sub infix:«<=>»(Mu $, Mu $?) is pure       { * }
+proto sub infix:«<=>»(Mu $, Mu $?) is pure { * }
 multi sub infix:«<=>»(\a, \b)  { a.Real <=> b.Real }
 
-proto sub infix:<==>(Mu $?, Mu $?) is pure  { * }
+proto sub infix:<==>(Mu $?, Mu $?) is pure { * }
 multi sub infix:<==>($?)        { Bool::True }
 multi sub infix:<==>(\a, \b)   { a.Numeric == b.Numeric }
 
+proto sub infix:<≅>(Mu $?, Mu $?, *%) { * }  # note, can't be pure due to dynvar
+multi sub infix:<≅>($?)        { Bool::True }
+multi sub infix:<≅>(\a, \b, :$tolerance = $*TOLERANCE)    {
+    # If operands are non-0, scale the tolerance to the larger of the abs values.
+    # We test b first since $value ≅ 0 is the usual idiom and falsifies faster.
+    if b && a && $tolerance {
+        abs(a - b) < (a.abs max b.abs) * $tolerance;
+    }
+    else {  # interpret tolerance as absolute
+        abs(a.Num - b.Num) < $tolerance;
+    }
+}
+sub infix:<=~=>(|c) { infix:<≅>(|c) }
+
 proto sub infix:<!=>(Mu $?, Mu $?) is pure  { * }
 multi sub infix:<!=>($?)        { Bool::True }
 multi sub infix:<!=>(Mu \a, Mu \b)   { not a == b }
@@ -274,12 +301,12 @@ multi sub infix:<+^>($x)         { $x }
 multi sub infix:<+^>($x, $y)     { $x.Numeric.Int +^ $y.Numeric.Int }
 
 proto sub infix:«+<»(Mu $?, Mu $?) is pure { * }
-multi sub infix:«+<»()           { fail "No zero-arg meaning for infix:«+<»"; }
+multi sub infix:«+<»() { Failure.new("No zero-arg meaning for infix:«+<»") }
 multi sub infix:«+<»($x)         { $x }
 multi sub infix:«+<»($x,$y)      { $x.Numeric.Int +< $y.Numeric.Int }
 
 proto sub infix:«+>»(Mu $?, Mu $?) is pure { * }
-multi sub infix:«+>»()           { fail "No zero-arg meaning for infix:«+>»"; }
+multi sub infix:«+>»() { Failure.new("No zero-arg meaning for infix:«+>»") }
 multi sub infix:«+>»($x)         { $x }
 multi sub infix:«+>»($x,$y)      { $x.Numeric.Int +> $y.Numeric.Int }
 
