@@ -4,14 +4,20 @@ my class Rat is Cool does Rational[Int, Int] {
     method FatRat(Rat:D: Real $?) { FatRat.new($!numerator, $!denominator); }
     method Range(Rat:U:) { Range.new(-Inf,Inf) }
     multi method perl(Rat:D:) {
-        my $d = $!denominator;
-        return $!numerator ~ '.0' if $d == 1;
-        unless $d == 0 {
-            $d div= 5 while $d %% 5;
-            $d div= 2 while $d %% 2;
-            self.REDUCE-ME;
+        if $!denominator == 1 {
+            $!numerator ~ '.0'
+        }
+        else {
+            my $d = $!denominator;
+            unless $d == 0 {
+                $d = $d div 5 while $d %% 5;
+                $d = $d div 2 while $d %% 2;
+                self.REDUCE-ME;
+            }
+            $d == 1
+              ?? self.base(10,*)
+              !! '<' ~ $!numerator ~ '/' ~ $!denominator ~ '>'
         }
-        ($d == 1) ?? self.base(10,*) !! '<' ~ $!numerator ~ '/' ~ $!denominator ~ '>';
     }
 }
 
@@ -19,8 +25,8 @@ my class FatRat is Cool does Rational[Int, Int] {
     method FatRat(FatRat:D: Real $?) { self }
     method Rat   (FatRat:D: Real $?) {
         $!denominator < $UINT64_UPPER
-            ?? Rat.new($!numerator, $!denominator)
-            !! fail "Cannot convert from FatRat to Rat because denominator is too big";
+          ?? Rat.new($!numerator, $!denominator)
+          !! Failure.new("Cannot convert from FatRat to Rat because denominator is too big")
     }
     multi method perl(FatRat:D:) {
         "FatRat.new($!numerator, $!denominator)";
@@ -204,12 +210,12 @@ multi sub infix:<%>(Rational \a, Rational \b) {
 multi sub infix:<**>(Rational \a, Int \b) {
     b >= 0
         ?? DIVIDE_NUMBERS
-            a.numerator ** b,
-            a.denominator ** b,
+            (a.numerator ** b // fail (a.numerator.abs > a.denominator ?? X::Numeric::Overflow !! X::Numeric::Underflow).new),
+            a.denominator ** b,  # we presume it likely already blew up on the numerator
             a,
             b
         !! DIVIDE_NUMBERS
-            a.denominator ** -b,
+            (a.denominator ** -b // fail (a.numerator.abs < a.denominator ?? X::Numeric::Overflow !! X::Numeric::Underflow).new),
             a.numerator ** -b,
             a,
             b
