@@ -58,9 +58,7 @@ my class Routine { # declared in BOOTSTRAP
         $perl
     }
 
-    method soft() {
-        Mu
-    }
+    method soft( --> Nil ) { }
 
     method wrap(&wrapper) {
         my class WrapHandle {
@@ -125,16 +123,4 @@ my class Routine { # declared in BOOTSTRAP
     }
 }
 
-multi sub trait_mod:<is>(Routine $r, :$cached!) {
-    my %cache;
-    nqp::bindattr_i($r, Routine, '$!onlystar', 0 )
-      if $r.onlystar; # disable optimization
-    $r.wrap(-> |c {
-        my $key := c.gist;
-        %cache.EXISTS-KEY($key)
-          ?? %cache{$key}
-          !! (%cache{$key} := callsame);
-    });
-}
-
 # vim: ft=perl6 expandtab sw=4
