@@ -1,7 +1,7 @@
 my role Setty does QuantHash {
     has %!elems; # key.WHICH => key
 
-    submethod BUILD (:%!elems)  { }
+    submethod BUILD(:%!elems --> Nil)  { }
     method default(--> Bool) { False }
 
     multi method keys(Setty:D:) {
@@ -22,14 +22,14 @@ my role Setty does QuantHash {
 
     method elems(Setty:D: --> Int) { %!elems.elems }
     method total(Setty:D: --> Int) { %!elems.elems }
-    method minpairs() { self.pairs }
-    method maxpairs() { self.pairs }
+    multi method minpairs(Setty:D:) { self.pairs }
+    multi method maxpairs(Setty:D:) { self.pairs }
     multi method Bool(Setty:D:) { %!elems.Bool }
 
     multi method hash(Setty:D: --> Hash) {
-        my %e;
-        %e{$_} = True for %!elems.values;
-        %e;
+        my \e = Hash.^parameterize(Bool, Any).new;
+        e{$_} = True for %!elems.values;
+        e;
     }
 
     multi method new(Setty: +@args) {
@@ -50,10 +50,14 @@ my role Setty does QuantHash {
         self.bless(:elems(%e));
     }
 
-    method ACCEPTS($other) {
-        self.defined
-          ?? $other (<=) self && self (<=) $other
-          !! $other.^does(self);
+    multi method ACCEPTS(Setty:U: $other) {
+        $other.^does(self)
+    }
+    multi method ACCEPTS(Setty:D: Seq:D \seq) {
+        self.ACCEPTS(seq.list)
+    }
+    multi method ACCEPTS(Setty:D: $other) {
+        $other (<=) self && self (<=) $other
     }
 
     multi method Str(Setty:D $ : --> Str) { ~ %!elems.values }
