@@ -27,7 +27,7 @@ my class Signature { # declared in BOOTSTRAP
         while @spos {
             my $s;
             my $t;
-            last unless $t=@tpos.shift;
+            last unless @tpos && ($t = @tpos.shift);
             $s=@spos.shift;
             if $s.slurpy or $s.capture {
                 @spos=();
@@ -58,12 +58,14 @@ my class Signature { # declared in BOOTSTRAP
         }
 
         my $here=$sclass{True}.SetHash;
-        my $hasslurpy=($sclass{True} // ()).grep({.slurpy}).Bool;
+        my $hasslurpy=($sclass{True} // ()).grep({.slurpy});
+        $here{@$hasslurpy} :delete;
+        $hasslurpy .= Bool;
         for flat @($tclass{True} // ()) -> $other {
             my $this;
 
             if $other.slurpy {
-                return False if any($here.keys) ~~ { !(.type =:= Mu) };
+                return False if any($here.keys) ~~ -> Any $_ { !(.type =:= Mu) };
                 return $hasslurpy;
             }
             if $this=$here.keys.grep( -> $t { $other ~~ $t }) {
@@ -73,6 +75,7 @@ my class Signature { # declared in BOOTSTRAP
                 return False unless $hasslurpy;
             }
         }
+        return False unless self.returns =:= $topic.returns;
         True;
     }
 
@@ -89,7 +92,8 @@ my class Signature { # declared in BOOTSTRAP
             nqp::clone($!params));
     }
 
-    method !gistperl(Signature:D: $perl, Mu:U :$elide-type = Mu) {
+    method !gistperl(Signature:D: $perl, Mu:U :$elide-type = Mu,
+                     :&where = -> $ { 'where { ... }' } ) {
         # Opening.
         my $text = $perl ?? ':(' !! '(';
 
@@ -102,7 +106,9 @@ my class Signature { # declared in BOOTSTRAP
 
             my $sep = '';
             for @params.kv -> $i, $param {
-                $text ~= $sep ~ $param.perl(:$elide-type);
+                my $parmstr = $param.perl(:$elide-type, :&where);
+                return Nil without $parmstr;
+                $text ~= $sep ~ $parmstr;
                 $text .= subst(/' $'$/,'') unless $perl;
                 $sep = $param.multi-invocant && !@params[$i+1].?multi-invocant
                   ?? ';; '
@@ -130,6 +136,73 @@ my class Signature { # declared in BOOTSTRAP
     method returns() { $!returns }
 }
 
-multi sub infix:<eqv>(Signature $a, Signature $b) { $a.perl eq $b.perl }
+multi sub infix:<eqv>(Signature \a, Signature \b) {
+
+    # we're us
+    return True if a =:= b;
+
+    # different container type
+    return False unless a.WHAT =:= b.WHAT;
+
+    # arity or count mismatch
+    return False if a.arity != b.arity || a.count != b.count;
+
+    # different number of parameters or no parameters
+    my $ap := nqp::getattr(a.params,List,'$!reified');
+    my $bp := nqp::getattr(b.params,List,'$!reified');
+    my int $elems = nqp::elems($ap);
+    return False if nqp::isne_i($elems,nqp::elems($bp));
+    return True unless $elems;
+
+    # compare all positionals
+    my int $i = -1;
+    Nil
+      while nqp::islt_i(++$i,$elems)
+        && nqp::atpos($ap,$i) eqv nqp::atpos($bp,$i);
+
+    # not all matching positionals
+    if nqp::islt_i($i,$elems) {
+
+        # not all same and different number of positionals
+        return False
+          if (!nqp::atpos($ap,$i).named || !nqp::atpos($bp,$i).named);
+
+        # create lookup table
+        my int $j = $i = $i - 1;
+        my $lookup := nqp::hash;
+        while nqp::islt_i(++$j,$elems) {
+            my $p  := nqp::atpos($ap,$j);
+            my $nn := nqp::getattr($p,Parameter,'$!named_names');
+            my str $key =
+              nqp::isnull($nn) ?? '' !! nqp::elems($nn) ?? nqp::atpos($nn,0) !! '';
+            die "Found named parameter '{
+              nqp::chars($key) ?? $key !! '(unnamed)'
+            }' twice in signature {a.perl}: {$p.perl} vs {nqp::atkey($lookup,$key).perl}"
+              if nqp::existskey($lookup,$key);
+            nqp::bindkey($lookup,$key,$p);
+        }
+
+        # named variable mismatch
+        while nqp::islt_i(++$i,$elems) {
+            my $p  := nqp::atpos($bp,$i);
+            my $nn := nqp::getattr($p,Parameter,'$!named_names');
+            my str $key = nqp::defined($nn) && nqp::elems($nn)
+              ?? nqp::atpos($nn,0)
+              !! '';
+
+            # named param doesn't exist in other or is not equivalent
+            return False
+              unless nqp::existskey($lookup,$key)
+                && $p eqv nqp::atkey($lookup,$key);
+        }
+    }
+
+    # it's a match
+    True
+}
+
+Perl6::Metamodel::Configuration.set_multi_sig_comparator(
+    -> \a, \b { a.signature eqv b.signature }
+);
 
 # vim: ft=perl6 expandtab sw=4
