@@ -3,8 +3,7 @@
 # Generates list of positions to index into the array at. Takes all those
 # before something lazy is encountered and eagerly reifies them. If there
 # are any lazy things in the slice, then we lazily consider those, but will
-# truncate at the first one that is out of range. We have a special case for
-# Range, which will auto-truncate even though not lazy.  The optional
+# truncate at the first one that is out of range. The optional
 # :$eagerize will be called if Whatever/WhateverCode is encountered or if
 # clipping of lazy indices is enacted.  It should return the number of
 # elements of the array if called with Whatever, or do something EXISTS-POSish
@@ -67,13 +66,6 @@ multi sub POSITIONS(\SELF, \pos, Callable :$eagerize = -> $idx {
     }
     pos-list
 }
-multi sub POSITIONS(\SELF, Range \pos) {
-    # Can be more clever here and look at range endpoints, as an optimization.
-    pos.map(-> Int() $i {
-        last unless SELF.EXISTS-POS($i);
-        $i
-    }).cache;
-}
 
 proto sub postcircumfix:<[ ]>(|) is nodal { * }
 
@@ -252,20 +244,15 @@ multi sub postcircumfix:<[ ]>(\SELF, Iterable:D \pos, Mu \val ) is raw {
 
         # Extract the values/containers which will be assigned to, in case
         # reifying the rhs does crazy things like splicing SELF.
-        my int $p = 0;
-        for indices {
-            nqp::bindpos($target,$p,SELF[$_]);
-            $p = $p + 1;
-        }
+        my int $p = -1;
+        nqp::bindpos($target,++$p,SELF[$_]) for indices;
 
         rvlist.EXISTS-POS($p);
         my \rviter := rvlist.iterator;
-        $p = 0;
+        $p = -1;
         my $elems = nqp::elems($target);
-        while $p < $elems {
-            nqp::atpos($target,$p) = rviter.pull-one;
-            $p = $p + 1;
-        }
+        nqp::atpos($target,$p) = rviter.pull-one
+          while nqp::islt_i(++$p,$elems);
         $list
     }
     else { # The assumption for now is this must be Iterable
@@ -300,25 +287,39 @@ multi sub postcircumfix:<[ ]>(\SELF, Iterable:D \pos, :$BIND!) is raw {
     X::Bind::Slice.new(type => SELF.WHAT).throw;
 }
 multi sub postcircumfix:<[ ]>(\SELF, Iterable:D \pos, :$SINK!, *%other) is raw {
-   SLICE_MORE_LIST( SELF, POSITIONS(SELF, pos), (:$SINK), %other );
+    nqp::iscont(pos)
+        ?? SLICE_ONE_LIST( SELF, pos.Int, (:$SINK), %other )
+        !! SLICE_MORE_LIST( SELF, POSITIONS(SELF, pos), (:$SINK), %other )
 }
 multi sub postcircumfix:<[ ]>(\SELF, Iterable:D \pos,:$delete!,*%other) is raw {
-   SLICE_MORE_LIST( SELF, POSITIONS(SELF, pos), (:$delete), %other );
+    nqp::iscont(pos)
+        ?? SLICE_ONE_LIST( SELF, pos.Int, (:$delete), %other )
+        !! SLICE_MORE_LIST( SELF, POSITIONS(SELF, pos), (:$delete), %other )
 }
 multi sub postcircumfix:<[ ]>(\SELF, Iterable:D \pos,:$exists!,*%other) is raw {
-   SLICE_MORE_LIST( SELF, POSITIONS(SELF, pos), (:$exists), %other );
+    nqp::iscont(pos)
+        ?? SLICE_ONE_LIST( SELF, pos.Int, (:$exists), %other )
+        !! SLICE_MORE_LIST( SELF, POSITIONS(SELF, pos), (:$exists), %other )
 }
 multi sub postcircumfix:<[ ]>(\SELF, Iterable:D \pos, :$kv!, *%other) is raw {
-   SLICE_MORE_LIST( SELF, POSITIONS(SELF, pos), (:$kv), %other );
+    nqp::iscont(pos)
+        ?? SLICE_ONE_LIST( SELF, pos.Int, (:$kv), %other )
+        !! SLICE_MORE_LIST( SELF, POSITIONS(SELF, pos), (:$kv), %other )
 }
 multi sub postcircumfix:<[ ]>(\SELF, Iterable:D \pos, :$p!, *%other) is raw {
-   SLICE_MORE_LIST( SELF, POSITIONS(SELF, pos), (:$p), %other );
+    nqp::iscont(pos)
+        ?? SLICE_ONE_LIST( SELF, pos.Int, (:$p), %other )
+        !! SLICE_MORE_LIST( SELF, POSITIONS(SELF, pos), (:$p), %other )
 }
 multi sub postcircumfix:<[ ]>(\SELF, Iterable:D \pos, :$k!, *%other) is raw {
-   SLICE_MORE_LIST( SELF, POSITIONS(SELF, pos), (:$k), %other );
+    nqp::iscont(pos)
+        ?? SLICE_ONE_LIST( SELF, pos.Int, (:$k), %other )
+        !! SLICE_MORE_LIST( SELF, POSITIONS(SELF, pos), (:$k), %other )
 }
 multi sub postcircumfix:<[ ]>(\SELF, Iterable:D \pos, :$v!, *%other) is raw {
-   SLICE_MORE_LIST( SELF, POSITIONS(SELF, pos), (:$v), %other );
+    nqp::iscont(pos)
+        ?? SLICE_ONE_LIST( SELF, pos.Int, (:$v), %other )
+        !! SLICE_MORE_LIST( SELF, POSITIONS(SELF, pos), (:$v), %other )
 }
 
 # @a[->{}]
@@ -507,30 +508,39 @@ sub MD-ARRAY-SLICE(\SELF, @indices) is raw {
 }
 
 multi sub postcircumfix:<[; ]>(\SELF, @indices) is raw {
-    my int $n = @indices.elems;
-    my int $i = 0;
-    my $all-ints := True;
-    while $i < $n {
-        $all-ints := False unless nqp::istype(@indices.AT-POS($i), Int);
-        $i = $i + 1;
-    }
-    $all-ints
-        ?? SELF.AT-POS(|@indices)
-        !! MD-ARRAY-SLICE(SELF, @indices)
+    my int $elems = @indices.elems;
+    my $indices := nqp::getattr(@indices,List,'$!reified');
+    my int $i = -1;
+
+    return MD-ARRAY-SLICE(SELF,@indices)
+      unless nqp::istype(nqp::atpos($indices,$i), Int)
+      while nqp::islt_i(++$i,$elems);
+
+    SELF.AT-POS(|@indices)
+}
+
+multi sub postcircumfix:<[; ]>(\SELF, @indices, Mu \assignee) is raw {
+    my int $elems = @indices.elems;
+    my $indices := nqp::getattr(@indices,List,'$!reified');
+    my int $i = -1;
+
+    return MD-ARRAY-SLICE(SELF, @indices) = assignee
+      unless nqp::istype(nqp::atpos($indices,$i), Int)
+      while nqp::islt_i(++$i,$elems);
+
+    SELF.ASSIGN-POS(|@indices, assignee)
 }
 
 multi sub postcircumfix:<[; ]>(\SELF, @indices, :$exists!) is raw {
     if $exists {
-        my int $n = @indices.elems;
-        my int $i = 0;
-        my $all-ints := True;
-        while $i < $n {
-            $all-ints := False unless nqp::istype(@indices.AT-POS($i), Int);
-            $i = $i + 1;
-        }
-        $all-ints
-            ?? SELF.EXISTS-POS(|@indices)
-            !! X::NYI.new(feature => ':exists on multi-dimensional slices')
+        my int $elems = @indices.elems;
+        my $indices := nqp::getattr(@indices,List,'$!reified');
+        my int $i = -1;
+        fail X::NYI.new(feature => ':exists on multi-dimensional slices')
+          unless nqp::istype(nqp::atpos($indices,$i), Int)
+          while nqp::islt_i(++$i,$elems);
+
+        SELF.EXISTS-POS(|@indices)
     }
     else {
         SELF[@indices]
@@ -539,16 +549,14 @@ multi sub postcircumfix:<[; ]>(\SELF, @indices, :$exists!) is raw {
 
 multi sub postcircumfix:<[; ]>(\SELF, @indices, :$delete!) is raw {
     if $delete {
-        my int $n = @indices.elems;
-        my int $i = 0;
-        my $all-ints := True;
-        while $i < $n {
-            $all-ints := False unless nqp::istype(@indices.AT-POS($i), Int);
-            $i = $i + 1;
-        }
-        $all-ints
-            ?? SELF.DELETE-POS(|@indices)
-            !! X::NYI.new(feature => ':delete on multi-dimensional slices')
+        my int $elems = @indices.elems;
+        my $indices := nqp::getattr(@indices,List,'$!reified');
+        my int $i = -1;
+        fail X::NYI.new(feature => ':delete on multi-dimensional slices')
+          unless nqp::istype(nqp::atpos($indices,$i), Int)
+          while nqp::islt_i(++$i,$elems);
+
+        SELF.DELETE-POS(|@indices)
     }
     else {
         SELF[@indices]
@@ -556,16 +564,14 @@ multi sub postcircumfix:<[; ]>(\SELF, @indices, :$delete!) is raw {
 }
 
 multi sub postcircumfix:<[; ]>(\SELF, @indices, :$BIND!) is raw {
-    my int $n = @indices.elems;
-    my int $i = 0;
-    my $all-ints := True;
-    while $i < $n {
-        $all-ints := False unless nqp::istype(@indices.AT-POS($i), Int);
-        $i = $i + 1;
-    }
-    $all-ints
-        ?? SELF.BIND-POS(|@indices, $BIND)
-        !! X::Bind::Slice.new(type => SELF.WHAT).throw;
+    my int $elems = @indices.elems;
+    my $indices := nqp::getattr(@indices,List,'$!reified');
+    my int $i = -1;
+    X::Bind::Slice.new(type => SELF.WHAT).throw
+      unless nqp::istype(nqp::atpos($indices,$i), Int)
+      while nqp::islt_i(++$i,$elems);
+
+    SELF.BIND-POS(|@indices, $BIND)
 }
 
 # vim: ft=perl6 expandtab sw=4
