@@ -195,7 +195,9 @@ sub SEQUENCE(\left, Mu \right, :$exclude_end) {
         }
         else {
             my $badseq;
-            my ($a, $b, $c);
+            my $a;
+            my $b;
+            my $c;
             unless $code.defined {
                 take @tail.shift while @tail.elems > 3;
                 $a = @tail[0];
@@ -254,7 +256,7 @@ sub SEQUENCE(\left, Mu \right, :$exclude_end) {
                 my $ab = $b - $a;
                 if $ab == $c - $b {
                     if $ab != 0 || nqp::istype($a,Real) && nqp::istype($b,Real) && nqp::istype($c,Real) {
-                        if nqp::istype($endpoint, Real) and nqp::isconcrete($endpoint) {
+                        if nqp::istype($endpoint, Real) and not nqp::istype($endpoint, Bool) and nqp::isconcrete($endpoint) {
                             if $ab > 0 {
                                 $stop = 1 if $a > $endpoint;
                                 $code = -> $x {
@@ -284,7 +286,7 @@ sub SEQUENCE(\left, Mu \right, :$exclude_end) {
                     $ab = $b / $a;
                     if $ab == $c / $b {
                         $ab = $ab.Int if nqp::istype($ab,Rat) && $ab.denominator == 1;
-                        if nqp::istype($endpoint, Real) and nqp::isconcrete($endpoint) {
+                        if nqp::istype($endpoint, Real) and not nqp::istype($endpoint, Bool) and nqp::isconcrete($endpoint) {
                             if $ab > 0 {
                                 if $ab > 1  {
                                     $stop = 1 if $a > $endpoint;
@@ -328,7 +330,7 @@ sub SEQUENCE(\left, Mu \right, :$exclude_end) {
             elsif @tail.elems == 2 {
                 my $ab = $b - $a;
                 if $ab != 0 || nqp::istype($a,Real) && nqp::istype($b,Real) {
-                    if nqp::istype($endpoint, Real) and nqp::isconcrete($endpoint) {
+                    if nqp::istype($endpoint, Real) and not nqp::istype($endpoint, Bool) and nqp::isconcrete($endpoint) {
                         if $ab > 0 {
                             $stop = 1 if $a > $endpoint;
                             $code = -> $x {
@@ -359,7 +361,7 @@ sub SEQUENCE(\left, Mu \right, :$exclude_end) {
                 if nqp::istype($endpoint,Code) or not nqp::isconcrete($endpoint) {
                     $code = { $^x.succ }
                 }
-                elsif nqp::istype($endpoint, Real) and nqp::istype($a, Real) {
+                elsif nqp::istype($endpoint, Real) and not nqp::istype($endpoint, Bool) and nqp::istype($a, Real) {
                     if $a < $endpoint {
                         $code = -> $x {
                             my $new = $x.succ;
@@ -564,56 +566,57 @@ sub INDIRECT_NAME_LOOKUP($root, *@chunks) is raw {
     $thing;
 }
 
-sub REQUIRE_IMPORT($package-name, *@syms) {
-    my $package = CALLER::OUR::($package-name);
-    my $who     = $package.WHO;
-    unless $who.EXISTS-KEY('EXPORT') {
-        die "Trying to import symbols @syms.join(', ') from '$package-name', but it does not export anything";
-    }
-    $who := $who<EXPORT>.WHO<DEFAULT>.WHO;
+sub REQUIRE_IMPORT($compunit, *@syms) {
+    my $handle := $compunit.handle;
+    my $DEFAULT := $handle.export-package()<DEFAULT>.WHO;
+    my $GLOBALish := $handle.globalish-package.WHO;
     my @missing;
+    # Set the runtime values for compile time stub symbols
     for @syms {
-        unless $who.EXISTS-KEY($_) {
+        unless $DEFAULT.EXISTS-KEY($_) {
             @missing.push: $_;
             next;
         }
-        OUTER::CALLER::{$_} := $who{$_};
+        OUTER::CALLER::{$_} := $DEFAULT{$_};
     }
     if @missing {
-        X::Import::MissingSymbols.new(:from($package-name), :@missing).throw;
+        X::Import::MissingSymbols.new(:from($compunit.short-name), :@missing).throw;
     }
-    $package
+    # Merge GLOBAL from compunit.
+    GLOBAL::.merge-symbols($GLOBALish);
+    Nil;
 }
-sub infix:<andthen>(**@a) {
-    return Bool::True unless @a;
-    my Mu $current := @a.shift;
-    for @a {
+
+sub infix:<andthen>(+a) {
+    my $ai := a.iterator;
+    my Mu $current := $ai.pull-one;
+    return Bool::True if $current =:= IterationEnd;
+    until ($_ := $ai.pull-one) =:= IterationEnd {
         return Empty unless $current.defined;
-        # Have to check Callable till we get tailthunky lists
         $current := $_ ~~ Callable
             ?? (.count ?? $_($current) !! $_())
             !! $_;
     }
     $current;
 }
-sub infix:<notandthen>(**@a) {
-    return Bool::True unless @a;
-    my Mu $current := @a.shift;
-    for @a {
+sub infix:<notandthen>(+a) {
+    my $ai := a.iterator;
+    my Mu $current := $ai.pull-one;
+    return Bool::True if $current =:= IterationEnd;
+    until ($_ := $ai.pull-one) =:= IterationEnd {
         return Empty if $current.defined;
-        # Have to check Callable till we get tailthunky lists
         $current := $_ ~~ Callable
             ?? (.count ?? $_($current) !! $_())
             !! $_;
     }
     $current;
 }
-sub infix:<orelse>(**@a) {
-    return Nil unless @a;
-    my Mu $current := @a.shift;
-    for @a {
+sub infix:<orelse>(+a) {
+    my $ai := a.iterator;
+    my Mu $current := $ai.pull-one;
+    return Nil if $current =:= IterationEnd;
+    until ($_ := $ai.pull-one) =:= IterationEnd {
         return $current if $current.defined;
-        # Have to check Callable till we get tailthunky lists
         $current := $_ ~~ Callable
             ?? (.count ?? $_($current) !! $_())
             !! $_;
@@ -628,7 +631,7 @@ multi sub trait_mod:<is>(Routine $r, Str :$equiv!) {
         my \nm ='&' ~ nqp::substr($r.name, 0, $i+1) ~ '<' ~ nqp::escape($equiv) ~ '>';
         trait_mod:<is>($r, equiv => ::(nm));
         return;
-    } 
+    }
     die "Routine given to equiv does not appear to be an operator";;
 }
 
@@ -637,7 +640,7 @@ multi sub trait_mod:<is>(Routine $r, Str :$tighter!) {
         my \nm ='&' ~ nqp::substr($r.name, 0, $i+1) ~ '<' ~ nqp::escape($tighter) ~ '>';
         trait_mod:<is>($r, tighter => ::(nm));
         return;
-    } 
+    }
     die "Routine given to tighter does not appear to be an operator";;
 }
 
@@ -646,7 +649,7 @@ multi sub trait_mod:<is>(Routine $r, Str :$looser!) {
         my \nm ='&' ~ nqp::substr($r.name, 0, $i+1) ~ '<' ~ nqp::escape($looser) ~ '>';
         trait_mod:<is>($r, looser => ::(nm));
         return;
-    } 
+    }
     die "Routine given to looser does not appear to be an operator";;
 }
 
@@ -656,4 +659,44 @@ multi sub infix:<∘> (&f) { &f }
 multi sub infix:<∘> (&f, &g --> Block) { (&f).count > 1 ?? -> |args { f |g |args } !! -> |args { f g |args } }
 my &infix:<o> := &infix:<∘>;
 
+# needs native arrays
+sub permutations(Int() $n) {
+    return ((),) if $n < 1;
+
+    Seq.new(
+        class :: does Iterator {
+            # See:  L<https://en.wikipedia.org/wiki/Permutation#Generation_in_lexicographic_order>
+            has int $!n;
+            has     @!a;
+            submethod BUILD(:$n --> Nil) { $!n = $n } # cannot set native in sig
+            #method is-lazy { True }
+            method pull-one {
+                return (@!a = ^$!n).List unless @!a;
+
+                # stop if there is only one value
+                # ( only needed because $k starts as 0 in this case )
+                return IterationEnd if $!n == 1;
+
+                # Find the largest index k such that a[k] < a[k + 1].
+                # If no such index exists, the permutation is the last permutation.
+                my int $k = @!a.end - 1;
+                $k-- or return IterationEnd until @!a[$k] < @!a[$k + 1];
+
+                # Find the largest index l greater than k such that a[k] < a[l].
+                my int $l = @!a.end;
+                $l-- until @!a[$k] < @!a[$l];
+                # use L<https://en.wikipedia.org/wiki/XOR_swap_algorithm>
+                # @!a[$k, $l].=reverse
+                (@!a[$k] +^= @!a[$l]) +^= @!a[$l] +^= @!a[$k];
+
+                # @!a[$k+1 .. @!a.end].=reverse;
+                $l = $!n;
+                (@!a[$k] +^= @!a[$l]) +^= @!a[$l] +^= @!a[$k] until ++$k >= --$l;
+                @!a.List;
+            }
+            method count-only { [*] 1 .. $!n }
+        }.new(:$n)
+    );
+}
+
 # vim: ft=perl6 expandtab sw=4
