@@ -7,6 +7,7 @@ my class X::Import::MissingSymbols   { ... }
 my class X::Redeclaration { ... }
 my class X::Inheritance::SelfInherit { ... }
 my class X::Comp::Trait::Unknown { ... }
+my class X::Experimental { ... }
 my class Pod::Block::Declarator { ... }
 
 proto sub trait_mod:<is>(|) { * }
@@ -101,16 +102,22 @@ multi sub trait_mod:<is>(Attribute:D $attr, :$trailing_docs!) {
 }
 
 multi sub trait_mod:<is>(Routine:D $r, |c ) {
-    X::Comp::Trait::Unknown.new(
-      file       => $?FILE,
-      line       => $?LINE,
-      type       => 'is',
-      subtype    => c.hash.keys[0],
-      declaring  => ' ' ~ lc( $r.^name ),
-      highexpect => ('rw raw hidden-from-backtrace hidden-from-USAGE',
-                     'cached pure default DEPRECATED inlinable nodal',
-                     'prec equiv tighter looser assoc leading_docs trailing_docs' ),
-    ).throw;
+    my $subtype = c.hash.keys[0];
+    $subtype eq 'cached'
+      ?? X::Experimental.new(
+        feature => "the 'is cached' trait",
+        use     => "cached",
+        ).throw
+      !! X::Comp::Trait::Unknown.new(
+        file       => $?FILE,
+        line       => $?LINE,
+        type       => 'is',
+        subtype    => $subtype,
+        declaring  => ' ' ~ lc( $r.^name ),
+        highexpect => ('rw raw hidden-from-backtrace hidden-from-USAGE',
+                       'pure default DEPRECATED inlinable nodal',
+                       'prec equiv tighter looser assoc leading_docs trailing_docs' ),
+        ).throw;
 }
 multi sub trait_mod:<is>(Routine:D $r, :$rw!) {
     $r.set_rw();
@@ -338,11 +345,6 @@ multi sub trait_mod:<returns>(Routine:D $target, Mu:U $type) {
     $target.^mixin(Callable.^parameterize($type))
 }
 
-proto sub trait_mod:<as>(|) { * }
-multi sub trait_mod:<as>(Parameter:D $param, $type) {
-    $param.set_coercion($type);
-}
-
 proto sub trait_mod:<handles>(|) { * }
 multi sub trait_mod:<handles>(Attribute:D $target, $thunk) {
     $target does role {
