@@ -4,6 +4,17 @@
 #define DLLEXPORT extern
 #endif
 
+#if defined(_MSC_VER) && _MSC_VER < 1800
+#  define bool unsigned char
+#  define true  1
+#  define false 0
+#  define __bool_true_false_are_defined 1
+#else
+#  include <stdbool.h>
+#endif
+
+#include <stddef.h>
+
 typedef struct {
     char  foo1;
     int   foo2;
@@ -60,3 +71,11 @@ DLLEXPORT int SizeofLng() {
 DLLEXPORT int SizeofPtr() {
     return sizeof(void *);
 }
+
+DLLEXPORT int SizeofBool() {
+    return sizeof(bool);
+}
+
+DLLEXPORT int SizeofSizeT() {
+    return sizeof(size_t);
+}
