@@ -1,4 +1,4 @@
-use v6;
+use v6.c;
 
 our %timing;
 
@@ -28,7 +28,7 @@ my class ProfiledGrammarHOW is Metamodel::GrammarHOW {
 }
 
 sub get-timing is export { %timing }
-sub reset-timing is export { %timing = {} }
+sub reset-timing is export { %timing = () }
 
 my module EXPORTHOW { }
 EXPORTHOW.WHO.<grammar> = ProfiledGrammarHOW;
